import asyncio
import contextvars
import functools
import logging
import tempfile
from io import SEEK_SET, StringIO

try:
    import importlib.metadata as metadata
except ImportError as e:
    # User is probably on Python<3.8
    try:
        import importlib_metadata as metadata
    except ImportError:
        # User needs to install importlib_metadata
        raise ImportError(
            f"importlib_metadata is not installed in Python<3.8. Please install it with "
            f"'pip install importlib_metadata'. {e}"
        )
import inspect
import json
import os
import re
import subprocess
import sys
import threading

from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Callable, Optional, Type, Union

import pexpect

from runhouse.constants import RH_LOGFILE_PATH
from runhouse.logger import get_logger, init_logger

logger = get_logger(__name__)


####################################################################################################
# Python package utilities
####################################################################################################
def find_locally_installed_version(package_name: str) -> Optional[str]:
    try:
        return metadata.version(package_name)
    except metadata.PackageNotFoundError:
        return None


def get_local_install_path(package_name: str) -> Optional[str]:
    distribution = metadata.distribution(package_name)
    direct_url_json = distribution.read_text("direct_url.json")
    if direct_url_json:
        # File URL starts with file://
        try:
            url = json.loads(direct_url_json).get("url", None)
            if url:
                if url.startswith("file://"):
                    return url[len("file://") :]
        except json.JSONDecodeError:
            return None


def is_python_package_string(s: str) -> bool:
    return isinstance(s, str) and re.match(r"^[a-zA-Z0-9\._-]+$", s) is not None


####################################################################################################
# Simple running utility
####################################################################################################
def run_with_logs(cmd: str, **kwargs):
    """Runs a command and prints the output to sys.stdout.
    We can't just pipe to sys.stdout, and when in a `call` method
    we overwrite sys.stdout with a multi-logger to a file and stdout.

    Args:
        cmd: The command to run.
        kwargs: Keyword arguments to pass to subprocess.Popen.

    Returns:
        The returncode of the command.
    """
    require_outputs = kwargs.pop("require_outputs", False)
    stream_logs = kwargs.pop("stream_logs", True)

    p = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        shell=True,
        **kwargs,
    )

    out = ""
    if stream_logs:
        while True:
            line = p.stdout.readline()
            if line == "" and p.poll() is not None:
                break
            sys.stdout.write(line)
            sys.stdout.flush()
            if require_outputs:
                out += line

    stdout, stderr = p.communicate()

    if require_outputs:
        stdout = stdout or out
        return p.returncode, stdout, stderr

    return p.returncode


####################################################################################################
# Module discovery and import logic
####################################################################################################


def _find_directory_containing_any_file(dir_path, files, searched_dirs=None):
    if Path(dir_path) == Path.home() or dir_path == Path("/"):
        return None

    if any(Path(dir_path, file).exists() for file in files):
        return str(dir_path)

    searched_dirs.add(dir_path)
    parent_path = Path(dir_path).parent
    if parent_path in searched_dirs:
        return None
    return _find_directory_containing_any_file(
        parent_path, files, searched_dirs=searched_dirs
    )


def locate_working_dir(start_dir=None):
    if start_dir is None:
        start_dir = os.getcwd()

    # Search first for anything that represents a Python package
    target_files = [
        ".git",
        "setup.py",
        "setup.cfg",
        "pyproject.toml",
        "requirements.txt",
    ]

    dir_with_target = _find_directory_containing_any_file(
        start_dir, target_files, searched_dirs=set()
    )

    if dir_with_target is None:
        dir_with_target = _find_directory_containing_any_file(
            start_dir, ["rh"], searched_dirs=set()
        )

    return dir_with_target if dir_with_target is not None else start_dir


def extract_module_path(raw_cls_or_fn: Union[Type, Callable]):
    py_module = inspect.getmodule(raw_cls_or_fn)

    # Need to resolve in case just filename is given
    module_path = (
        str(Path(inspect.getfile(py_module)).resolve())
        if hasattr(py_module, "__file__")
        else None
    )

    return module_path


def get_module_import_info(raw_cls_or_fn: Union[Type, Callable]):
    """
    Given a class or function in Python, get all the information needed to import it in another Python process.
    """

    # Background on all these dunders: https://docs.python.org/3/reference/import.html
    py_module = inspect.getmodule(raw_cls_or_fn)

    # Need to resolve in case just filename is given
    module_path = extract_module_path(raw_cls_or_fn)

    # TODO better way of detecting if in a notebook or interactive Python env
    if not module_path or module_path.endswith("ipynb"):
        # The only time __file__ wouldn't be present is if the function is defined in an interactive
        # interpreter or a notebook. We can't import on the server in that case, so we need to cloudpickle
        # the fn to send it over. The __call__ function will serialize the function if we return it this way.
        # This is a short-term hack.
        # return None, "notebook", raw_fn.__name__
        root_path = os.getcwd()
        module_name = "notebook"
        cls_or_fn_name = raw_cls_or_fn.__name__
    else:
        root_path = os.path.dirname(module_path)
        module_name = inspect.getmodulename(module_path)
        # TODO __qualname__ doesn't work when fn is aliased funnily, like torch.sum
        cls_or_fn_name = getattr(raw_cls_or_fn, "__qualname__", raw_cls_or_fn.__name__)

        # Adapted from https://github.com/modal-labs/modal-client/blob/main/modal/_function_utils.py#L94
        if getattr(py_module, "__package__", None):
            module_path = os.path.abspath(py_module.__file__)
            package_paths = [
                os.path.abspath(p) for p in __import__(py_module.__package__).__path__
            ]
            base_dirs = [
                base_dir
                for base_dir in package_paths
                if os.path.commonpath((base_dir, module_path)) == base_dir
            ]

            if len(base_dirs) != 1:
                logger.debug(f"Module files: {module_path}")
                logger.debug(f"Package paths: {package_paths}")
                logger.debug(f"Base dirs: {base_dirs}")
                raise Exception("Wasn't able to find the package directory!")
            root_path = os.path.dirname(base_dirs[0])
            module_name = py_module.__spec__.name

    return root_path, module_name, cls_or_fn_name


####################################################################################################
# Run command with password
####################################################################################################


def run_command_with_password_login(
    command: str, password: str, stream_logs: bool = True
):
    command_run = pexpect.spawn(command, encoding="utf-8", timeout=None)
    if stream_logs:
        # FYI This will print a ton of of stuff to stdout
        command_run.logfile_read = sys.stdout

    # If CommandRunner uses the control path, the password may not be requested
    next_line = command_run.expect(["assword:", pexpect.EOF])
    if next_line == 0:
        command_run.sendline(password)
        command_run.expect(pexpect.EOF)
    command_run.close()

    return command_run


####################################################################################################
# Async helpers
####################################################################################################


def thread_coroutine(coroutine, context=None):
    # Copy contextvars from the parent thread to the new thread
    if context is not None:
        for var, value in context.items():
            var.set(value)

    # Technically, event loop logic is not threadsafe. However, this event loop is only in this thread.
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        # The loop runs only for the duration of the thread
        return loop.run_until_complete(coroutine)
    finally:
        # We don't need to do asyncio.set_event_loop(None) since the thread will just end completely
        loop.close()


# We should minimize calls to this since each one will start a new thread.
# Technically we should not have many threads running async logic at once, however, the calling thread
# will actually block until the async logic that is spawned in the other thread is done.
def sync_function(coroutine_func):
    @functools.wraps(coroutine_func)
    def wrapper(*args, **kwargs):
        # Better API than using threading.Thread, since we just need the thread temporarily
        # and the resources are cleaned up
        with ThreadPoolExecutor() as executor:
            future = executor.submit(
                thread_coroutine,
                coroutine_func(*args, **kwargs),
                contextvars.copy_context(),
            )
            return future.result()

    return wrapper


async def arun_in_thread(method_to_run, *args, **kwargs):
    def _run_sync_fn_with_context(context_to_set, sync_fn, method_args, method_kwargs):
        for var, value in context_to_set.items():
            var.set(value)

        return sync_fn(*method_args, **method_kwargs)

    with ThreadPoolExecutor() as executor:
        return await asyncio.get_event_loop().run_in_executor(
            executor,
            functools.partial(
                _run_sync_fn_with_context,
                context_to_set=contextvars.copy_context(),
                sync_fn=method_to_run,
                method_args=args,
                method_kwargs=kwargs,
            ),
        )


####################################################################################################
# Misc helpers
####################################################################################################


def get_pid():
    import os

    return os.getpid()


def get_node_ip():
    import socket

    return socket.gethostbyname(socket.gethostname())


class ThreadWithException(threading.Thread):
    def run(self):
        self._exc = None
        try:
            super().run()
        except Exception as e:
            self._exc = e

    def join(self, timeout=None):
        super().join(timeout=timeout)
        if self._exc:
            raise self._exc


def client_call_wrapper(client, system, client_method_name, *args, **kwargs):
    from runhouse.resources.hardware import Cluster

    if system and isinstance(system, Cluster) and not system.on_this_cluster():
        return system.call_client_method(client_method_name, *args, **kwargs)
    method = getattr(client, client_method_name)
    return method(*args, **kwargs)


####################################################################################################
# Logging redirection
####################################################################################################
class StreamTee(object):
    def __init__(self, instream, outstreams):
        self.instream = instream
        self.outstreams = outstreams

    def write(self, message):
        self.instream.write(message)
        for stream in self.outstreams:
            if message:
                stream.write(message)
                # We flush here to ensure that the logs are written to the file immediately
                # see https://github.com/run-house/runhouse/pull/724
                stream.flush()

    def writelines(self, lines):
        self.instream.writelines(lines)
        for stream in self.outstreams:
            stream.writelines(lines)
            stream.flush()

    def flush(self):
        self.instream.flush()
        for stream in self.outstreams:
            stream.flush()

    def __getattr__(self, item):
        # Needed in case someone calls a method on instream, such as Ray calling sys.stdout.istty()
        return getattr(self.instream, item)


class capture_stdout:
    """Context manager for capturing stdout to a file, list, or stream, while still printing to stdout."""

    def __init__(self, output=None):
        self.output = output
        self._stream = None

    def __enter__(self):
        if self.output is None:
            self.output = StringIO()

        if isinstance(self.output, str):
            self._stream = open(self.output, "w")
        else:
            self._stream = self.output
        sys.stdout = StreamTee(sys.stdout, [self])
        sys.stderr = StreamTee(sys.stderr, [self])
        return self

    def write(self, message):
        self._stream.write(message)

    def flush(self):
        self._stream.flush()

    @property
    def stream(self):
        if isinstance(self.output, str):
            return open(self.output, "r")
        return self._stream

    def list(self):
        if isinstance(self.output, str):
            return self.stream.readlines()
        return (self.stream.getvalue() or "").splitlines()

    def __str__(self):
        return self.stream.getvalue()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(sys.stdout, "instream"):
            sys.stdout = sys.stdout.instream
        if hasattr(sys.stderr, "instream"):
            sys.stderr = sys.stderr.instream
        self._stream.close()
        return False


class LogToFolder:
    def __init__(self, name: str):
        self.name = name
        self.directory = self._base_local_folder_path(name)
        # We do exist_ok=True here because generator runs are separate calls to the same directory.
        os.makedirs(self.directory, exist_ok=True)
        self.logger = None
        self.handler = None

    def __enter__(self):
        # TODO fix the fact that we keep appending and then stream back the full file
        sys.stdout = StreamTee(sys.stdout, [Path(self._stdout_path).open(mode="a")])
        sys.stderr = StreamTee(sys.stderr, [Path(self._stderr_path).open(mode="a")])

        self.logger = logging.getLogger()
        init_logger(self.logger)
        self.handler = logging.FileHandler(self._stdout_path)
        self.logger.addHandler(self.handler)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        # Flush stdout and stderr
        sys.stdout.flush()
        sys.stderr.flush()

        # Restore stdout and stderr
        if hasattr(sys.stdout, "instream"):
            sys.stdout = sys.stdout.instream
        if hasattr(sys.stderr, "instream"):
            sys.stderr = sys.stderr.instream

        # Close the file handler
        self.handler.close()
        self.logger.removeHandler(self.handler)
        init_logger(self.logger)

        # return False to propagate any exception that occurred inside the with block
        return False

    @property
    def _stdout_path(self) -> str:
        """Path to the stdout file for the Run."""
        return self._path_to_file_by_ext(ext=".out")

    @property
    def _stderr_path(self) -> str:
        """Path to the stderr file for the Run."""
        return self._path_to_file_by_ext(ext=".err")

    @staticmethod
    def _base_local_folder_path(name: str):
        """Path to the base folder for this Run on a local system."""
        return f"{RH_LOGFILE_PATH}/{name}"

    @staticmethod
    def _filter_files_by_ext(files: list, ext: str):
        return list(filter(lambda x: x.endswith(ext), files))

    def _find_file_path_by_ext(self, ext: str) -> Union[str, None]:
        """Get the file path by provided extension. Needed when loading the stdout and stderr files associated
        with a particular run."""
        try:
            # List all files in self.directory
            folder_contents = os.listdir(self.directory)
        except FileNotFoundError:
            return None

        files_with_ext = self._filter_files_by_ext(folder_contents, ext)
        if not files_with_ext:
            # No .out / .err file already created in the logs folder for this Run
            return None

        # Return the most recent file with this extension
        return f"{self.directory}/{files_with_ext[0]}"

    def _path_to_file_by_ext(self, ext: str) -> str:
        """Path the file for the Run saved on the system for a provided extension (ex: ``.out`` or ``.err``)."""
        existing_file = self._find_file_path_by_ext(ext=ext)
        if existing_file:
            # If file already exists in file (ex: with function on a Ray cluster this will already be
            # generated for us)
            return existing_file

        path_to_ext = f"{self.directory}/{self.name}" + ext
        return path_to_ext


class SuppressStd(object):
    """Context to capture stderr and stdout at C-level."""

    def __init__(self, outfile=None):
        self.orig_stdout_fileno = sys.__stdout__.fileno()
        self.orig_stderr_fileno = sys.__stderr__.fileno()
        self.output = None

    def __enter__(self):
        # Redirect the stdout/stderr fd to temp file
        self.orig_stdout_dup = os.dup(self.orig_stdout_fileno)
        self.orig_stderr_dup = os.dup(self.orig_stderr_fileno)
        self.tfile = tempfile.TemporaryFile(mode="w+b")
        os.dup2(self.tfile.fileno(), self.orig_stdout_fileno)
        os.dup2(self.tfile.fileno(), self.orig_stderr_fileno)

        # Store the stdout object and replace it by the temp file.
        self.stdout_obj = sys.stdout
        self.stderr_obj = sys.stderr
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        return self

    def __exit__(self, exc_class, value, traceback):

        # Make sure to flush stdout
        print(flush=True)

        # Restore the stdout/stderr object.
        sys.stdout = self.stdout_obj
        sys.stderr = self.stderr_obj

        # Close capture file handle
        os.close(self.orig_stdout_fileno)
        os.close(self.orig_stderr_fileno)

        # Restore original stderr and stdout
        os.dup2(self.orig_stdout_dup, self.orig_stdout_fileno)
        os.dup2(self.orig_stderr_dup, self.orig_stderr_fileno)

        # Close duplicate file handle.
        os.close(self.orig_stdout_dup)
        os.close(self.orig_stderr_dup)

        # Copy contents of temporary file to the given stream
        self.tfile.flush()
        self.tfile.seek(0, SEEK_SET)
        self.output = self.tfile.read().decode()
        self.tfile.close()


####################################################################################################
# Name generation
####################################################################################################


def generate_default_name(prefix: str = None, precision: str = "s", sep="_") -> str:
    """Name of the Run's parent folder which contains the Run's data (config, stdout, stderr, etc).
    If a name is provided, prepend that to the current timestamp to complete the folder name."""
    if precision == "d":
        timestamp_key = f"{datetime.now().strftime('%Y%m%d')}"
    elif precision == "s":
        timestamp_key = f"{datetime.now().strftime(f'%Y%m%d{sep}%H%M%S')}"
    elif precision == "ms":
        timestamp_key = f"{datetime.now().strftime(f'%Y%m%d{sep}%H%M%S_%f')}"
    if prefix is None:
        return timestamp_key
    return f"{prefix}{sep}{timestamp_key}"


####################################################################################################
# Logger utils
####################################################################################################
class ColoredFormatter:
    COLORS = {
        "black": "\u001b[30m",
        "red": "\u001b[31m",
        "green": "\u001b[32m",
        "yellow": "\u001b[33m",
        "blue": "\u001b[34m",
        "magenta": "\u001b[35m",
        "cyan": "\u001b[36m",
        "white": "\u001b[37m",
        "reset": "\u001b[0m",
    }

    @classmethod
    def get_color(cls, color: str):
        return cls.COLORS.get(color, "")

    # TODO: This method is a temp solution, until we'll update logging architecture. Remove once logging is cleaned up.
    @classmethod
    def format_log(cls, text):
        ansi_escape = re.compile(r"(?:\x1B[@-_][0-?]*[ -/]*[@-~])")
        return ansi_escape.sub("", text)


class ClusterLogsFormatter:
    def __init__(self, system):
        self.system = system
        self._display_title = False

    def format(self, output_type):
        from runhouse import Resource
        from runhouse.servers.http.http_utils import OutputType

        system_color = ColoredFormatter.get_color("cyan")
        reset_color = ColoredFormatter.get_color("reset")

        prettify_logs = output_type in [
            OutputType.STDOUT,
            OutputType.EXCEPTION,
            OutputType.STDERR,
        ]

        if (
            isinstance(self.system, Resource)
            and prettify_logs
            and not self._display_title
        ):
            # Display the system name before subsequent logs only once
            system_name = self.system.name
            dotted_line = "-" * len(system_name)
            print(dotted_line)
            print(f"{system_color}{system_name}{reset_color}")
            print(dotted_line)

            # Only display the system name once
            self._display_title = True

        return system_color, reset_color


def create_local_dir(path: Union[str, Path]):
    full_path = os.path.expanduser(path) if isinstance(path, str) else path.expanduser()
    Path(full_path).parent.mkdir(parents=True, exist_ok=True)
    return full_path


####################################################################################################
# Status collection utils
####################################################################################################
class ServletType(str, Enum):
    env = "env"
    cluster = "cluster"


def get_gpu_usage(collected_gpus_info: dict, servlet_type: ServletType):

    gpus_indices = list(collected_gpus_info.keys())

    # how we retrieve total_gpu_memory:
    # 1. getting the first gpu usage of the first gpu un the gpus list
    # 2. getting the first gpu_info dictionary of the specific gpu (we collected the gpu info over time)
    # 3. get total_memory value (it is the same across all envs)
    total_gpu_memory = collected_gpus_info[gpus_indices[0]][0].get("total_memory")
    total_used_memory, gpu_utilization_percent, free_memory = 0, 0, 0

    if servlet_type == ServletType.cluster:
        free_memory = collected_gpus_info[gpus_indices[0]][-1].get(
            "free_memory"
        )  # getting the latest free_memory value collected.

    for gpu_index in gpus_indices:
        collected_gpu_info = collected_gpus_info.get(gpu_index)
        sum_used_memery = sum(
            [gpu_info.get("used_memory") for gpu_info in collected_gpu_info]
        )
        total_used_memory = sum_used_memery / len(collected_gpu_info)  # average

        if servlet_type == ServletType.cluster:
            sum_cpu_util = sum(
                [gpu_info.get("utilization_percent") for gpu_info in collected_gpu_info]
            )
            gpu_utilization_percent = sum_cpu_util / len(collected_gpu_info)  # average

    total_used_memory = total_used_memory / len(gpus_indices)

    gpu_usage = {"total_memory": total_gpu_memory, "used_memory": total_used_memory}

    if servlet_type == ServletType.cluster:
        gpu_utilization_percent = round(gpu_utilization_percent / len(gpus_indices), 2)
        gpu_usage["free_memory"] = free_memory
        gpu_usage["gpu_count"] = len(gpus_indices)
        gpu_usage["utilization_percent"] = gpu_utilization_percent

    return gpu_usage



import asyncio
import time

import pytest
import requests

import runhouse as rh

from runhouse.globals import rns_client
from runhouse.resources.hardware.utils import ResourceServerStatus

import tests.test_resources.test_clusters.test_cluster
from tests.utils import friend_account


def set_autostop_from_on_cluster_via_ah(mins):
    ah = rh.servers.autostop_helper.AutostopHelper()

    asyncio.run(ah.set_autostop(mins))


def get_auotstop_from_on_cluster():
    ah = rh.servers.autostop_helper.AutostopHelper()

    return asyncio.run(ah.get_autostop())


def get_last_active_time_from_on_cluster():
    ah = rh.servers.autostop_helper.AutostopHelper()

    return asyncio.run(ah.get_last_active_time())


def register_activity_from_on_cluster():
    ah = rh.servers.autostop_helper.AutostopHelper()

    asyncio.run(ah.set_last_active_time_to_now())
    asyncio.run(ah.register_activity_if_needed())


def set_autostop_from_on_cluster_via_cluster_obj(mins):
    rh.here.autostop_mins = mins


def set_autostop_from_on_cluster_via_cluster_keep_warm():
    rh.here.keep_warm()


def torch_exists():
    try:
        import torch

        torch.rand(4)
        return True
    except ImportError:
        return False


class TestOnDemandCluster(tests.test_resources.test_clusters.test_cluster.TestCluster):

    MAP_FIXTURES = {"resource": "cluster"}

    UNIT = {"cluster": []}
    LOCAL = {"cluster": []}
    MINIMAL = {
        "cluster": [
            "ondemand_aws_docker_cluster",
            "ondemand_gcp_cluster",
            "ondemand_k8s_cluster",
        ]
    }
    RELEASE = {
        "cluster": [
            "ondemand_aws_docker_cluster",
            "ondemand_gcp_cluster",
            "ondemand_aws_https_cluster_with_auth",
            "ondemand_k8s_cluster",
            "ondemand_k8s_docker_cluster",
        ]
    }
    MAXIMAL = {
        "cluster": [
            "ondemand_aws_docker_cluster",
            "ondemand_gcp_cluster",
            "ondemand_k8s_cluster",
            "ondemand_k8s_docker_cluster",
            "ondemand_aws_https_cluster_with_auth",
            "v100_gpu_cluster",
            "k80_gpu_cluster",
            "a10g_gpu_cluster",
            "static_cpu_pwd_cluster",
            "multinode_cpu_docker_conda_cluster",
            "multinode_gpu_cluster",
        ]
    }

    @pytest.mark.level("minimal")
    def test_restart_does_not_change_config_yaml(self, cluster):
        assert cluster.up_if_not()
        config_yaml_res = cluster.run("cat ~/.rh/config.yaml")
        assert config_yaml_res[0][0] == 0
        config_yaml_content = config_yaml_res[0][1]

        cluster.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        with friend_account():
            cluster.restart_server()
            config_yaml_res_after_restart = cluster.run("cat ~/.rh/config.yaml")
            assert config_yaml_res_after_restart[0][0] == 0
            config_yaml_content_after_restart = config_yaml_res[0][1]
            assert config_yaml_content_after_restart == config_yaml_content

    @pytest.mark.level("minimal")
    def test_set_autostop(self, cluster):
        rh.env(
            working_dir="local:./", reqs=["pytest", "pandas"], name="autostop_env"
        ).to(cluster)

        get_autostop = rh.fn(get_auotstop_from_on_cluster).to(
            cluster, env="autostop_env"
        )
        # First check that the autostop is set to whatever the cluster set it to
        assert get_autostop() == cluster.autostop_mins
        original_autostop = cluster.autostop_mins

        set_autostop = rh.fn(set_autostop_from_on_cluster_via_ah).to(
            cluster, env="autostop_env"
        )
        set_autostop(5)
        assert get_autostop() == 5

        set_autostop_via_cluster_keep_warm = rh.fn(
            set_autostop_from_on_cluster_via_cluster_keep_warm
        ).to(cluster, env="autostop_env")
        set_autostop_via_cluster_keep_warm()
        assert get_autostop() == -1

        set_autostop_via_cluster_obj = rh.fn(
            set_autostop_from_on_cluster_via_cluster_obj
        ).to(cluster, env="autostop_env")
        # reset the autostop to the original value
        set_autostop_via_cluster_obj(original_autostop)
        assert get_autostop() == original_autostop

    @pytest.mark.level("minimal")
    def test_register_activity(self, cluster):
        rh.env(
            working_dir="local:./", reqs=["pytest", "pandas"], name="autostop_env"
        ).to(cluster)

        register_activity = rh.fn(register_activity_from_on_cluster).to(
            cluster, env="autostop_env"
        )
        get_last_active = rh.fn(get_last_active_time_from_on_cluster).to(
            cluster, env="autostop_env"
        )

        register_activity()
        # Check that last active is within the last 2 seconds
        assert get_last_active() > time.time() - 3

    # TODO add a way to manually trigger the status loop to check that activity
    #  is actually registered after a call
    # cluster.call("autostop_env", "config")
    # cluster.status()
    # assert get_last_active() > time.time() - 2

    # TODO add a way to manually trigger the status loop to check that activity
    #  is actually registered during a long running function
    # from .test_cluster import sleep_fn
    # sleep_remote = rh.fn(sleep_fn).to(cluster, env="autostop_env")
    # Thread(target=sleep_remote, args=(3,)).start()
    # time.sleep(2)
    # cluster.status()
    # # Check that last active is within the last second, so we know the activity wasn't just from the call itself
    # assert get_last_active() > time.time() - 1

    @pytest.mark.level("release")
    def test_cluster_ping_and_is_up(self, cluster):
        assert cluster._ping(retry=False)

        original_ips = cluster.ips

        cluster.address = None
        assert not cluster._ping(retry=False)

        if not (
            hasattr(cluster, "launched_properties")
            and cluster.launched_properties["cloud"] == "kubernetes"
        ):  # kubernetes does not use ips in command runner
            cluster.address = "00.00.000.11"
            assert not cluster._ping(retry=False)

        assert cluster._ping(retry=True)
        assert cluster.is_up()
        assert cluster.ips == original_ips

    @pytest.mark.level("release")
    def test_docker_container_reqs(self, ondemand_aws_docker_cluster):
        ret_code = ondemand_aws_docker_cluster.run("pip freeze | grep torch")[0][0]
        assert ret_code == 0

    @pytest.mark.level("release")
    def test_fn_to_docker_container(self, ondemand_aws_docker_cluster):
        remote_torch_exists = rh.function(torch_exists).to(ondemand_aws_docker_cluster)
        assert remote_torch_exists()

    ####################################################################################################
    # Status tests
    ####################################################################################################

    # TODO: Affects cluster state, causes other tests to fail with ssh connection errors
    @pytest.mark.skip()
    @pytest.mark.level("minimal")
    def test_set_status_after_teardown(self, cluster, mocker):
        mock_function = mocker.patch("sky.down")
        response = cluster.teardown()
        assert isinstance(response, int)
        assert (
            response == 200
        )  # that means that the call to post status endpoint in den was successful
        mock_function.assert_called_once()

        cluster_config = cluster.config()
        cluster_uri = rns_client.format_rns_address(cluster.rns_address)
        api_server_url = cluster_config.get("api_server_url", rns_client.api_server_url)
        cluster.teardown()
        get_status_data_resp = requests.get(
            f"{api_server_url}/resource/{cluster_uri}/cluster/status",
            headers=rns_client.request_headers(),
        )

        assert get_status_data_resp.status_code == 200
        # For UI displaying purposes, the cluster/status endpoint returns cluster status history.
        # The latest status info is the first element in the list returned by the endpoint.
        get_status_data = get_status_data_resp.json()["data"][0]
        assert get_status_data["resource_type"] == cluster_config.get("resource_type")
        assert get_status_data["status"] == ResourceServerStatus.terminated

        assert cluster.is_up()



import copy
import os
import threading
import time
import traceback
from functools import wraps
from typing import Any, Dict, Optional

import psutil
import pynvml

from runhouse import configs

from runhouse.constants import (
    DEFAULT_STATUS_CHECK_INTERVAL,
    GPU_COLLECTION_INTERVAL,
    MAX_GPU_INFO_LEN,
    REDUCED_GPU_INFO_LEN,
)

from runhouse.globals import obj_store
from runhouse.logger import get_logger

from runhouse.resources.hardware.utils import detect_cuda_version_or_cpu

from runhouse.servers.http.http_utils import (
    deserialize_data,
    handle_exception_response,
    OutputType,
    Response,
    serialize_data,
)
from runhouse.servers.obj_store import ClusterServletSetupOption

from runhouse.utils import (
    arun_in_thread,
    get_gpu_usage,
    get_node_ip,
    get_pid,
    ServletType,
)

logger = get_logger(__name__)


def error_handling_decorator(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        ctx = kwargs.pop("ctx", None)
        ctx_token = None
        if ctx:
            ctx_token = obj_store.set_ctx(**ctx)

        serialization = kwargs.get("serialization", None)
        if "data" in kwargs:
            serialized_data = kwargs.get("data", None)
            deserialized_data = deserialize_data(serialized_data, serialization)
            kwargs["data"] = deserialized_data

        # If serialization is None, then we have not called this from the server,
        # so we should return the result of the function directly, or raise
        # the exception if there is one, instead of returning a Response object.
        try:
            output = await func(*args, **kwargs)
            if serialization is None or serialization == "none":
                return output
            if output is not None:
                if kwargs.get("remote"):
                    return Response(
                        output_type=OutputType.CONFIG,
                        data=output,
                    )
                serialized_data = serialize_data(output, serialization)
                return Response(
                    output_type=OutputType.RESULT_SERIALIZED,
                    data=serialized_data,
                    serialization=serialization,
                )
            else:
                return Response(
                    output_type=OutputType.SUCCESS,
                )
        except Exception as e:
            return handle_exception_response(e, traceback.format_exc(), serialization)
        finally:
            if ctx_token:
                obj_store.unset_ctx(ctx_token)

    return wrapper


class EnvServlet:
    async def __init__(self, env_name: str, *args, **kwargs):
        self.env_name = env_name

        await obj_store.ainitialize(
            self.env_name,
            has_local_storage=True,
            setup_cluster_servlet=ClusterServletSetupOption.GET_OR_FAIL,
        )

        # Ray defaults to setting OMP_NUM_THREADS to 1, which unexpectedly limit parallelism in user programs.
        # We delete it by default, but if we find that the user explicitly set it to another value, we respect that.
        # This is really only a factor if the user set the value inside the VM or container, or inside the base_env
        # which a cluster was initialized with. If they set it inside the env constructor and the env was sent to the
        # cluster normally with .to, it will be set after this point.
        # TODO this had no effect when we did it below where we set CUDA_VISIBLE_DEVICES, so we may need to move that
        #  here and mirror the same behavior (setting it based on the detected gpus in the whole cluster may not work
        #  for multinode, but popping it may also break things, it needs to be tested).
        num_threads = os.environ.get("OMP_NUM_THREADS")
        if num_threads is not None and num_threads != "1":
            os.environ["OMP_NUM_THREADS"] = num_threads
        else:
            os.environ["OMP_NUM_THREADS"] = ""

        self.output_types = {}
        self.thread_ids = {}

        self.pid = get_pid()
        self.process = psutil.Process(pid=self.pid)

        self.gpu_metrics = None  # will be updated only if this is a gpu cluster.
        self.lock = (
            threading.Lock()
        )  # will be used when self.gpu_metrics will be updated by different threads.

        if detect_cuda_version_or_cpu() != "cpu":
            logger.debug("Creating _periodic_gpu_check thread.")
            collect_gpu_thread = threading.Thread(
                target=self._collect_env_gpu_usage, daemon=True
            )
            collect_gpu_thread.start()

    ##############################################################
    # Methods to disable or enable den auth
    ##############################################################
    async def aset_cluster_config(self, cluster_config: Dict[str, Any]):
        obj_store.cluster_config = cluster_config

    async def aset_cluster_config_value(self, key: str, value: Any):
        obj_store.cluster_config[key] = value

    ##############################################################
    # Methods decorated with a standardized error decorating handler
    # These catch exceptions and wrap the output in a Response object.
    # They also handle arbitrary serialization and deserialization.
    # NOTE: These need to take in "data" and "serialization" as arguments
    # even if unused, because they are used by the decorator
    ##############################################################
    @error_handling_decorator
    async def aput_resource_local(
        self,
        data: Any,  # This first comes in a serialized format which the decorator re-populates after deserializing
        serialization: Optional[str] = None,
    ):
        resource_config, state, dryrun = tuple(data)
        return await obj_store.aput_resource_local(resource_config, state, dryrun)

    @error_handling_decorator
    async def aput_local(
        self, key: Any, data: Any, serialization: Optional[str] = None
    ):
        return await obj_store.aput_local(key, data)

    @error_handling_decorator
    async def acall_local(
        self,
        key: Any,
        method_name: str = None,
        data: Any = None,
        serialization: Optional[str] = None,
        run_name: Optional[str] = None,
        stream_logs: bool = False,
        remote: bool = False,
        ctx: Optional[dict] = None,
    ):
        if data is not None:
            args, kwargs = data.get("args", []), data.get("kwargs", {})
        else:
            args, kwargs = [], {}

        return await obj_store.acall_local(
            key,
            method_name,
            run_name=run_name,
            stream_logs=stream_logs,
            remote=remote,
            *args,
            **kwargs,
        )

    @error_handling_decorator
    async def aget_local(
        self,
        key: Any,
        default: Optional[Any] = None,
        serialization: Optional[str] = None,
        remote: bool = False,
    ):
        return obj_store.get_local(key, default=default, remote=remote)

    ##############################################################
    # IPC methods for interacting with local object store only
    # These do not catch exceptions, and do not wrap the output
    # in a Response object.
    ##############################################################
    async def akeys_local(self):
        return obj_store.keys_local()

    async def arename_local(self, key: Any, new_key: Any):
        return await obj_store.arename_local(key, new_key)

    async def acontains_local(self, key: Any):
        return obj_store.contains_local(key)

    async def apop_local(self, key: Any, *args):
        return await obj_store.apop_local(key, *args)

    async def adelete_local(self, key: Any):
        return await obj_store.adelete_local(key)

    async def aclear_local(self):
        return await obj_store.aclear_local()

    def _get_env_cpu_usage(self, cluster_config: dict = None):

        total_memory = psutil.virtual_memory().total
        node_ip = get_node_ip()

        if not cluster_config.get("resource_subtype") == "Cluster":
            stable_internal_external_ips = cluster_config.get(
                "stable_internal_external_ips", []
            )
            for ips_set in stable_internal_external_ips:
                internal_ip, _ = ips_set[0], ips_set[1]
                if internal_ip == node_ip:
                    # head ip equals to cluster address equals to cluster.ips[0]
                    if ips_set[1] == cluster_config.get("ips")[0]:
                        node_index = 0
                    else:
                        node_index = stable_internal_external_ips.index(ips_set)
                    node_name = f"worker_{node_index}"
        else:
            # a case it is a BYO cluster, assume that first ip in the ips list is the head.
            ips = cluster_config.get("ips", [])
            if len(ips) == 1 or node_ip == ips[0]:
                node_index = 0
            else:
                node_index = ips.index(node_ip)
            node_name = f"worker_{node_index}"
        try:

            memory_size_bytes = self.process.memory_full_info().uss
            cpu_usage_percent = self.process.cpu_percent(interval=0)
            env_memory_usage = {
                "used_memory": memory_size_bytes,
                "utilization_percent": cpu_usage_percent,
                "total_memory": total_memory,
            }
        except psutil.NoSuchProcess:
            env_memory_usage = {}

        return (
            env_memory_usage,
            node_name,
            total_memory,
            self.pid,
            node_ip,
            node_index,
        )

    def _get_env_gpu_usage(self):
        # currently works correctly for a single node GPU. Multinode-clusters will be supported shortly.

        collected_gpus_info = copy.deepcopy(self.gpu_metrics)

        if collected_gpus_info is None or not collected_gpus_info[0]:
            return None

        return get_gpu_usage(
            collected_gpus_info=collected_gpus_info, servlet_type=ServletType.env
        )

    def _collect_env_gpu_usage(self):
        """periodically collects env gpu usage"""

        pynvml.nvmlInit()  # init nvidia ml info collection

        while True:
            try:
                gpu_count = pynvml.nvmlDeviceGetCount()
                with self.lock:
                    if not self.gpu_metrics:
                        self.gpu_metrics: Dict[int, list[Dict[str, int]]] = {
                            device: [] for device in range(gpu_count)
                        }

                    for gpu_index in range(gpu_count):
                        handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_index)
                        processes = pynvml.nvmlDeviceGetComputeRunningProcesses_v3(
                            handle
                        )
                        memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
                        if processes:
                            for p in processes:
                                if p.pid == self.pid:
                                    used_memory = p.usedGpuMemory  # in bytes
                                    total_memory = memory_info.total  # in bytes
                                    current_gpu_metrics: list[
                                        Dict[str, int]
                                    ] = self.gpu_metrics[gpu_index]
                                    # to reduce cluster memory usage (we are saving the gpu_usage info on the cluster),
                                    # we save only the most updated gpu usage. If for some reason the size of updated_gpu_info is
                                    # too big, we remove the older gpu usage info.
                                    # This is relevant when using cluster.status() directly and not relying on status being sent to den.
                                    if len(current_gpu_metrics) + 1 > MAX_GPU_INFO_LEN:
                                        current_gpu_metrics = current_gpu_metrics[
                                            REDUCED_GPU_INFO_LEN:
                                        ]
                                    current_gpu_metrics.append(
                                        {
                                            "used_memory": used_memory,
                                            "total_memory": total_memory,
                                        }
                                    )
                                    self.gpu_metrics[gpu_index] = current_gpu_metrics
            except Exception as e:
                logger.error(str(e))
                pynvml.nvmlShutdown()
                break
            finally:
                # collects gpu usage every 5 seconds.
                time.sleep(GPU_COLLECTION_INTERVAL)

    def _status_local_helper(self):
        objects_in_env_servlet = obj_store.keys_with_info()
        cluster_config = obj_store.cluster_config

        (
            env_memory_usage,
            node_name,
            total_memory,
            env_servlet_pid,
            node_ip,
            node_index,
        ) = self._get_env_cpu_usage(cluster_config)

        # Try loading GPU data (if relevant)
        env_gpu_usage = (
            self._get_env_gpu_usage() if cluster_config.get("has_cuda", False) else {}
        )

        cluster_config = obj_store.cluster_config
        interval_size = cluster_config.get(
            "status_check_interval", DEFAULT_STATUS_CHECK_INTERVAL
        )

        # TODO: [sb]: once introduced, we could use ClusterServlet _cluster_periodic_thread_alive() to replace the
        #  'should_send_status_and_logs_to_den' logic below.
        # Only if one of these is true, do we actually need to get the status from each EnvServlet
        should_send_status_and_logs_to_den: bool = (
            configs.token is not None and interval_size != -1
        )

        # reset the gpu_info only if the current env_gpu collection will be sent to den. Otherwise, keep collecting it.
        if should_send_status_and_logs_to_den:
            with self.lock:
                self.gpu_metrics = None

        env_servlet_utilization_data = {
            "env_gpu_usage": env_gpu_usage,
            "node_ip": node_ip,
            "node_name": node_name,
            "node_index": node_index,
            "env_cpu_usage": env_memory_usage,
            "pid": env_servlet_pid,
        }

        return objects_in_env_servlet, env_servlet_utilization_data

    async def astatus_local(self):
        return await arun_in_thread(self._status_local_helper)



import json
import os
from pathlib import Path
from typing import Dict, Optional, Union

from runhouse.globals import rns_client
from runhouse.logger import get_logger
from runhouse.rns.utils.api import load_resp_content, read_resp_data


USER_ENDPOINT = "user/secret"


logger = get_logger(__name__)


def load_config(name: str, endpoint: str = USER_ENDPOINT):
    if "/" not in name:
        name = f"{rns_client.current_folder}/{name}"
    rns_address = rns_client.resolve_rns_path(name)
    if rns_address.startswith("/"):
        # Load via Resource API
        rns_config = rns_client.load_config(name=name)
        if not rns_config:
            raise ValueError(f"Secret {name} not found in Den.")

        # Load via Secrets API
        resource_uri = rns_client.resource_uri(name)
        secret_values = _load_vault_secret(resource_uri, endpoint)
        return {**rns_config, **{"values": secret_values}}

    # Load from local config
    return _load_local_config(name)


def _load_vault_secret(
    resource_uri: str,
    endpoint: str = USER_ENDPOINT,
    headers: Optional[Dict] = None,
):
    """Load secrets data from Vault for a particular resource URI. By default we allow for reloading shared secrets."""
    headers = headers or rns_client.request_headers()
    uri = f"{rns_client.api_server_url}/{endpoint}/{resource_uri}?shared=true"
    resp = rns_client.session.get(
        uri,
        headers=headers,
    )
    if resp.status_code != 200:
        raise Exception(
            f"Received [{resp.status_code}] from Den GET '{uri}': Failed to load secret from Vault: {load_resp_content(resp)}"
        )

    config = read_resp_data(resp)

    if len(config.keys()) == 1:
        vault_key = list(config.keys())[0]
        config = config[vault_key]
    if config.get("data", None):
        return config["data"]["values"]
    return config


def _delete_vault_secrets(
    resource_uri: str,
    endpoint: str = USER_ENDPOINT,
    headers: Optional[Dict] = None,
):
    headers = headers or rns_client.request_headers()
    uri = f"{rns_client.api_server_url}/{endpoint}/{resource_uri}"
    resp = rns_client.session.delete(
        uri,
        headers=headers,
    )
    if resp.status_code != 200:
        logger.error(
            f"Received [{resp.status_code}] from Den DELETE '{uri}': Failed to delete secrets from Vault: {load_resp_content(resp)}"
        )


def _load_local_config(name):
    if name.startswith("~") or name.startswith("^"):
        name = name[2:]
    config_path = os.path.expanduser(f"~/.rh/secrets/{name}.json")
    if not Path(config_path).exists():
        return None

    logger.info(f"Loading config from local file {config_path}")
    with open(config_path, "r") as f:
        try:
            config = json.load(f)
        except json.decoder.JSONDecodeError as e:
            logger.error(f"Error loading config from {config_path}: {e}")
            return None
    return config


def _is_matching_subset(existing_vals, new_vals):
    for key in new_vals:
        if key in existing_vals and existing_vals[key] != new_vals[key]:
            return False
    return True


def _check_file_for_mismatches(path, existing_vals, new_vals, overwrite):
    # Returns True if performs the check satisfactorily and not overriding the file.
    # Returns False if no existing vals exist or overwriting the file.
    if not existing_vals:
        return False
    if _is_matching_subset(existing_vals, new_vals):
        logger.info(f"Secrets already exist in {path}.")
        return True
    elif not overwrite:
        logger.warning(
            f"Path {path} already exists with a different set of values, "
            "and overwrite is set to False, so leaving the file as is. Please set overwrite "
            "to True or manually overwrite the secret file if you intend to do so."
        )
        return True
    return False


def setup_cluster_creds(ssh_creds: Union[Dict, str], resource_name: str):
    """
    This method creates an Secret corresponding to a cluster's SSH creds. If the passed values are paths to private
    and/or public keys, this method extracts the content of the files saved in those files, in order for them to
    be saved in Vault. (Currently if we just pass a path/to/ssh/key to the Secret constructor, the content of the file
    will not be saved to Vault. We need to pass the content itself.)

    Args:
       ssh_creds (Dict or str): the ssh credentials of the cluster, passed by the user, dict.
       resource_name (str): the name of the resource that the ssh secret is associated with.

    Returns:
       SSHSecret
    """
    import runhouse as rh
    from runhouse.resources.secrets import Secret
    from runhouse.resources.secrets.provider_secrets import ProviderSecret
    from runhouse.resources.secrets.provider_secrets.ssh_secret import SSHSecret

    if isinstance(ssh_creds, str):
        return Secret.from_name(name=ssh_creds)

    creds_keys = list(ssh_creds.keys())

    if len(creds_keys) == 1 and "ssh_private_key" in creds_keys:
        if Path(ssh_creds["ssh_private_key"]).expanduser().exists():
            values = SSHSecret.extract_secrets_from_path(path=ssh_creds["private_key"])
            values["ssh_private_key"] = ssh_creds["private_key"]
        else:
            # case where the user decides to pass the private key as text and not as path.
            raise ValueError(
                "SSH creds require both private and public key, but only private key was provided"
            )

    elif "ssh_private_key" in creds_keys and "ssh_public key" in creds_keys:
        private_key, public_key = (
            ssh_creds["ssh_private_key"],
            ssh_creds["ssh_public_key"],
        )
        private_key_path, public_key_path = (
            Path(private_key).expanduser(),
            Path(public_key).expanduser(),
        )
        if private_key_path.exists() and public_key_path.exists():
            values = SSHSecret.extract_secrets_from_path(path=private_key)
            values["ssh_private_key"], values["ssh_public_key"] = (
                private_key_path,
                public_key_path,
            )
        else:
            values = {"private_key": private_key, "public_key": public_key}

    elif "ssh_private_key" in creds_keys and "ssh_user" in creds_keys:
        private_key, username = ssh_creds["ssh_private_key"], ssh_creds["ssh_user"]
        if Path(private_key).expanduser().exists():
            private_key = SSHSecret.extract_secrets_from_path(path=private_key).get(
                "private_key"
            )
        if Path(username).expanduser().exists():
            username = ProviderSecret.extract_secrets_from_path(username)
        values = {
            "private_key": private_key,
            "ssh_user": username,
            "ssh_private_key": ssh_creds["ssh_private_key"],
        }

    elif "ssh_user" in creds_keys and "password" in creds_keys:
        password, username = ssh_creds["password"], ssh_creds["ssh_user"]
        if Path(password).expanduser().exists():
            password = ProviderSecret.extract_secrets_from_path(password)
        if Path(username).expanduser().exists():
            username = ProviderSecret.extract_secrets_from_path(username)
        values = {"password": password, "ssh_user": username}

    else:
        values = {}
        for k in creds_keys:
            v = ssh_creds[k]
            if Path(v).exists():
                v = ProviderSecret.extract_secrets_from_path(v)
            values.update({k: v})

    values_to_add = {k: ssh_creds[k] for k in ssh_creds if k not in values.keys()}
    values.update(values_to_add)

    new_secret = rh.secret(name=f"{resource_name}-ssh-secret", values=values)
    return new_secret



import asyncio
import contextlib
import json
import subprocess
import time
import warnings
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Any, Dict, List, Union

import requests

import rich.errors
import yaml

try:
    import sky
    from sky.backends import backend_utils
except ImportError:
    pass

from runhouse.constants import (
    DEFAULT_HTTP_PORT,
    DEFAULT_HTTPS_PORT,
    DEFAULT_SERVER_PORT,
    DOCKER_LOGIN_ENV_VARS,
    LOCAL_HOSTS,
)

from runhouse.globals import configs, obj_store, rns_client
from runhouse.logger import get_logger
from runhouse.resources.hardware.utils import (
    ResourceServerStatus,
    ServerConnectionType,
    up_cluster_helper,
)

from .cluster import Cluster

logger = get_logger(__name__)


class OnDemandCluster(Cluster):
    RESOURCE_TYPE = "cluster"
    RECONNECT_TIMEOUT = 5
    DEFAULT_KEYFILE = "~/.ssh/sky-key"

    def __init__(
        self,
        name,
        instance_type: str = None,
        num_instances: int = None,
        provider: str = None,
        default_env: "Env" = None,
        dryrun: bool = False,
        autostop_mins: int = None,
        use_spot: bool = False,
        image_id: str = None,
        memory: Union[int, str] = None,
        disk_size: Union[int, str] = None,
        open_ports: Union[int, str, List[int]] = None,
        server_host: int = None,
        server_port: int = None,
        server_connection_type: str = None,
        ssl_keyfile: str = None,
        ssl_certfile: str = None,
        domain: str = None,
        den_auth: bool = False,
        region: str = None,
        sky_kwargs: Dict = None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        On-demand `SkyPilot <https://github.com/skypilot-org/skypilot/>`__ Cluster.

        .. note::
            To build a cluster, please use the factory method :func:`cluster`.
        """
        super().__init__(
            name=name,
            default_env=default_env,
            server_host=server_host,
            server_port=server_port,
            server_connection_type=server_connection_type,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            domain=domain,
            den_auth=den_auth,
            dryrun=dryrun,
            **kwargs,
        )

        self.instance_type = instance_type
        self.num_instances = num_instances
        self.provider = provider or configs.get("default_provider")
        self._autostop_mins = (
            autostop_mins
            if autostop_mins is not None
            else configs.get("default_autostop")
        )

        self.open_ports = open_ports
        self.use_spot = use_spot if use_spot is not None else configs.get("use_spot")
        self.image_id = image_id
        self.region = region
        self.memory = memory
        self.disk_size = disk_size
        self.sky_kwargs = sky_kwargs or {}

        self.stable_internal_external_ips = kwargs.get(
            "stable_internal_external_ips", None
        )
        self.launched_properties = kwargs.get("launched_properties", {})
        self._docker_user = None

        # Checks if state info is in local sky db, populates if so.
        if not dryrun and not self.ips and not self.creds_values:
            # Cluster status is set to INIT in the Sky DB right after starting, so we need to refresh once
            self._update_from_sky_status(dryrun=False)

    @property
    def client(self):
        try:
            return super().client
        except ValueError as e:
            if not self.address:
                # Try loading in from local Sky DB
                self._update_from_sky_status(dryrun=True)
                if not self.address:
                    raise ValueError(
                        f"Could not determine address for ondemand cluster <{self.name}>. "
                        "Up the cluster with `cluster.up_if_not`."
                    )
                return super().client
            raise e

    @property
    def autostop_mins(self):
        return self._autostop_mins

    @autostop_mins.setter
    def autostop_mins(self, mins):
        self._autostop_mins = mins
        if self.on_this_cluster():
            obj_store.set_cluster_config_value("autostop_mins", mins)
        else:
            # if self.run_python(["import skypilot"])[0] != 0:
            #     raise ImportError(
            #         "Skypilot must be installed on the cluster in order to set autostop."
            #     )
            self.call_client_method("set_settings", {"autostop_mins": mins})
            sky.autostop(self.name, mins, down=True)

    @property
    def docker_user(self) -> str:
        if self._docker_user:
            return self._docker_user

        # TODO detect whether this is a k8s cluster properly, and handle the user setting / SSH properly
        #  (e.g. SkyPilot's new KubernetesCommandRunner)
        if not self.image_id or "docker:" not in self.image_id:
            return None

        if self.launched_properties["cloud"] == "kubernetes":
            return self.launched_properties.get(
                "docker_user", self.launched_properties.get("ssh_user", "root")
            )

        from runhouse.resources.hardware.sky_command_runner import get_docker_user

        if not self._creds:
            return
        self._docker_user = get_docker_user(self, self._creds.values)

        return self._docker_user

    def config(self, condensed=True):
        config = super().config(condensed)
        self.save_attrs_to_config(
            config,
            [
                "instance_type",
                "num_instances",
                "provider",
                "open_ports",
                "use_spot",
                "image_id",
                "region",
                "stable_internal_external_ips",
                "memory",
                "disk_size",
                "sky_kwargs",
                "launched_properties",
            ],
        )
        config["autostop_mins"] = self._autostop_mins
        return config

    def endpoint(self, external: bool = False):
        if not self.address or self.on_this_cluster():
            return None

        try:
            self.client.check_server()
        except ConnectionError:
            return None

        return super().endpoint(external)

    def _copy_sky_yaml_from_cluster(self, abs_yaml_path: str):
        if not Path(abs_yaml_path).exists():
            Path(abs_yaml_path).parent.mkdir(parents=True, exist_ok=True)
            self.rsync("~/.sky/sky_ray.yml", abs_yaml_path, up=False)

            # Save SSH info to the ~/.ssh/config
            ray_yaml = yaml.safe_load(open(abs_yaml_path, "r"))
            backend_utils.SSHConfigHelper.add_cluster(
                self.name, [self.address], ray_yaml["auth"]
            )

    @staticmethod
    def relative_yaml_path(yaml_path):
        if Path(yaml_path).is_absolute():
            yaml_path = "~/.sky/generated/" + Path(yaml_path).name
        return yaml_path

    def set_connection_defaults(self):
        if not self.server_connection_type:
            if self.ssl_keyfile or self.ssl_certfile:
                self.server_connection_type = ServerConnectionType.TLS
            else:
                self.server_connection_type = ServerConnectionType.SSH

        if self.server_port is None:
            if self.server_connection_type == ServerConnectionType.TLS:
                self.server_port = DEFAULT_HTTPS_PORT
            elif self.server_connection_type == ServerConnectionType.NONE:
                self.server_port = DEFAULT_HTTP_PORT
            else:
                self.server_port = DEFAULT_SERVER_PORT

        if (
            self.server_connection_type
            in [ServerConnectionType.TLS, ServerConnectionType.NONE]
            and self.server_host in LOCAL_HOSTS
        ):
            warnings.warn(
                f"Server connection type: {self.server_connection_type}, server host: {self.server_host}. "
                f"Note that this will require opening an SSH tunnel to forward traffic from"
                f" {self.server_host} to the server."
            )

        self.open_ports = (
            []
            if self.open_ports is None
            else [self.open_ports]
            if isinstance(self.open_ports, (int, str))
            else self.open_ports
        )

        if self.open_ports:
            self.open_ports = [str(p) for p in self.open_ports]
            if str(self.server_port) in self.open_ports:
                if (
                    self.server_connection_type
                    in [ServerConnectionType.TLS, ServerConnectionType.NONE]
                    and not self.den_auth
                ):
                    warnings.warn(
                        "Server is insecure and must be inside a VPC or have `den_auth` enabled to secure it."
                    )
            else:
                warnings.warn(
                    f"Server port {self.server_port} not included in open ports. Note you are responsible for opening "
                    f"the port or ensure you have access to it via a VPC."
                )
        else:
            # If using HTTP or HTTPS must enable traffic on the relevant port
            if self.server_connection_type in [
                ServerConnectionType.TLS,
                ServerConnectionType.NONE,
            ]:
                if self.server_port:
                    warnings.warn(
                        f"No open ports specified. Setting default port {self.server_port} to open."
                    )
                    self.open_ports = [str(self.server_port)]
                else:
                    warnings.warn(
                        f"No open ports specified. Make sure the relevant port is open. "
                        f"HTTPS default: {DEFAULT_HTTPS_PORT} and HTTP "
                        f"default: {DEFAULT_HTTP_PORT}."
                    )

    # ----------------- Launch/Lifecycle Methods -----------------

    def is_up(self) -> bool:
        """Whether the cluster is up.

        Example:
            >>> rh.ondemand_cluster("rh-cpu").is_up()
        """
        if self.on_this_cluster():
            return True
        return self._ping(retry=True)

    def _sky_status(self, refresh: bool = True, retry: bool = True):
        """
        Get status of Sky cluster.

        Return dict looks like:

        .. code-block::

            {'name': 'sky-cpunode-donny',
             'launched_at': 1662317201,
             'handle': ResourceHandle(
                          cluster_name=sky-cpunode-donny,
                          head_ip=54.211.97.164,
                          cluster_yaml=/Users/donny/.sky/generated/sky-cpunode-donny.yml,
                          launched_resources=1x AWS(m6i.2xlarge),
                          tpu_create_script=None,
                          tpu_delete_script=None),
             'last_use': 'sky cpunode',
             'status': <ClusterStatus.UP: 'UP'>,
             'autostop': -1,
             'metadata': {}}


        .. note:: For more information see SkyPilot's :code:`ResourceHandle` `class
        <https://github.com/skypilot-org/skypilot/blob/0c2b291b03abe486b521b40a3069195e56b62324/sky/backends/cloud_vm_ray_backend.py#L1457>`__.
        """
        if not sky.global_user_state.get_cluster_from_name(self.name):
            return None

        try:
            state = sky.status(cluster_names=[self.name], refresh=refresh)
        except rich.errors.LiveError as e:
            # We can't have more than one Live display at once, so if we've already launched one (e.g. the first
            # time we call status), we can retry without refreshing
            if not retry:
                raise e
            return self._sky_status(refresh=False, retry=False)

        # We still need to check if the cluster present in case the cluster went down and was removed from the DB
        if len(state) == 0:
            return None
        return state[0]

    @property
    def internal_ips(self):
        if not self.stable_internal_external_ips:
            self._update_from_sky_status()
        return [int_ip for int_ip, _ in self.stable_internal_external_ips]

    def _start_ray_workers(self, ray_port, env):
        # Find the internal IP corresponding to the public_head_ip and the rest are workers
        internal_head_ip = None
        worker_ips = []
        stable_internal_external_ips = self._sky_status()[
            "handle"
        ].stable_internal_external_ips
        for internal, external in stable_internal_external_ips:
            if external == self.address:
                internal_head_ip = internal
            else:
                # NOTE: Using external worker address here because we're running from local
                worker_ips.append(external)

        logger.debug(f"Internal head IP: {internal_head_ip}")

        for host in worker_ips:
            logger.info(
                f"Starting Ray on worker {host} with head node at {internal_head_ip}:{ray_port}."
            )
            self.run(
                commands=[
                    f"ray start --address={internal_head_ip}:{ray_port} --disable-usage-stats",
                ],
                node=host,
                env=env,
            )
        time.sleep(5)

    def _populate_connection_from_status_dict(self, cluster_dict: Dict[str, Any]):
        if cluster_dict and cluster_dict["status"].name in ["UP", "INIT"]:
            handle = cluster_dict["handle"]
            self.address = handle.head_ip
            self.stable_internal_external_ips = handle.stable_internal_external_ips
            if self.stable_internal_external_ips is None or self.address is None:
                raise ValueError(
                    "Sky's cluster status does not have the necessary information to connect to the cluster. Please check if the cluster is up via `sky status`. Consider bringing down the cluster with `sky down` if you are still having issues."
                )
            yaml_path = handle.cluster_yaml
            if Path(yaml_path).exists():
                ssh_values = backend_utils.ssh_credential_from_yaml(
                    yaml_path, ssh_user=handle.ssh_user
                )
                if not self.creds_values:
                    from runhouse.resources.secrets.utils import setup_cluster_creds

                    self._creds = setup_cluster_creds(ssh_values, self.name)

            # Add worker IPs if multi-node cluster - keep the head node as the first IP
            self.ips = [ext for _, ext in self.stable_internal_external_ips]

            launched_resource = handle.launched_resources
            cloud = str(launched_resource.cloud).lower()
            instance_type = launched_resource.instance_type
            region = launched_resource.region
            cost_per_hr = launched_resource.get_cost(60 * 60)
            disk_size = launched_resource.disk_size
            num_cpus = launched_resource.cpus

            self.launched_properties = {
                "cloud": cloud,
                "instance_type": instance_type,
                "region": region,
                "cost_per_hour": str(cost_per_hr),
                "disk_size": disk_size,
                "num_cpus": num_cpus,
            }
            if launched_resource.accelerators:
                self.launched_properties[
                    "accelerators"
                ] = launched_resource.accelerators
            if handle.ssh_user:
                self.launched_properties["ssh_user"] = handle.ssh_user
            if handle.docker_user:
                self.launched_properties["docker_user"] = handle.docker_user
            if cloud == "kubernetes":
                try:
                    import kubernetes

                    _, context = kubernetes.config.list_kube_config_contexts()
                    if "namespace" in context["context"]:
                        namespace = context["context"]["namespace"]
                    else:
                        namespace = "default"
                except:
                    namespace = "default"
                pod_name = f"{handle.cluster_name_on_cloud}-head"
                self.launched_properties["namespace"] = namespace
                self.launched_properties["pod_name"] = pod_name
        else:
            self.address = None
            self._creds = None
            self.stable_internal_external_ips = None
            self.launched_properties = {}

    def _update_from_sky_status(self, dryrun: bool = False):
        # Try to get the cluster status from SkyDB
        if self.is_shared:
            # If the cluster is shared can ignore, since the sky data will only be saved on the machine where
            # the cluster was initially upped
            return

        cluster_dict = self._sky_status(refresh=not dryrun)
        self._populate_connection_from_status_dict(cluster_dict)

    def get_instance_type(self):
        """Returns instance type of the cluster."""
        if self.instance_type and "--" in self.instance_type:  # K8s specific syntax
            return self.instance_type
        elif (
            self.instance_type
            and ":" not in self.instance_type
            and "CPU" not in self.instance_type
        ):
            return self.instance_type

        return None

    def accelerators(self):
        """Returns the acclerator type, or None if is a CPU."""
        if (
            self.instance_type
            and ":" in self.instance_type
            and "CPU" not in self.instance_type
        ):
            return self.instance_type

        return None

    def num_cpus(self):
        """Return the number of CPUs for a CPU cluster."""
        if (
            self.instance_type
            and ":" in self.instance_type
            and "CPU" in self.instance_type
        ):
            return self.instance_type.rsplit(":", 1)[1]

        return None

    async def a_up(self, capture_output: Union[bool, str] = True):
        """Up the cluster async in another process, so it can be parallelized and logs can be captured sanely.

        capture_output: If True, supress the output of the cluster creation process. If False, print the output
        normally. If a string, write the output to the file at that path.
        """

        with ProcessPoolExecutor() as executor:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(
                executor, up_cluster_helper, self, capture_output
            )
        return self

    async def a_up_if_not(self, capture_output: Union[bool, str] = True):
        if not self.is_up():
            # Don't store stale IPs
            self.ips = None
            await self.a_up(capture_output=capture_output)
        return self

    def up(self):
        """Up the cluster.

        Example:
            >>> rh.ondemand_cluster("rh-cpu").up()
        """
        if self.on_this_cluster():
            return self

        supported_providers = ["cheapest"] + list(sky.clouds.CLOUD_REGISTRY)
        if self.provider not in supported_providers:
            raise ValueError(
                f"Cluster provider {self.provider} not supported. Must be one {supported_providers} supported by SkyPilot."
            )

        task = sky.Task(num_nodes=self.num_instances)
        cloud_provider = (
            sky.clouds.CLOUD_REGISTRY.from_str(self.provider)
            if self.provider != "cheapest"
            else None
        )
        try:
            task.set_resources(
                sky.Resources(
                    # TODO: confirm if passing instance type in old way (without --) works when provider is k8s
                    cloud=cloud_provider,
                    instance_type=self.get_instance_type(),
                    accelerators=self.accelerators(),
                    cpus=self.num_cpus(),
                    memory=self.memory,
                    region=self.region or configs.get("default_region"),
                    disk_size=self.disk_size,
                    ports=self.open_ports,
                    image_id=self.image_id,
                    use_spot=self.use_spot,
                    **self.sky_kwargs.get("resources", {}),
                )
            )
            if self.image_id:
                import os

                docker_env_vars = {}
                for env_var in DOCKER_LOGIN_ENV_VARS:
                    if os.getenv(env_var):
                        docker_env_vars[env_var] = os.getenv(env_var)
                if docker_env_vars:
                    task.update_envs(docker_env_vars)
            sky.launch(
                task,
                cluster_name=self.name,
                idle_minutes_to_autostop=self._autostop_mins,
                down=True,
                **self.sky_kwargs.get("launch", {}),
            )
        # Make sure no args are passed both in sky_kwargs and as explicit args
        except TypeError as e:
            if "got multiple values for keyword argument" in str(e):
                raise TypeError(
                    f"{str(e)}. If argument is in `sky_kwargs`, it may need to be passed directly through the "
                    f"ondemand_cluster constructor (see `ondemand_cluster docs "
                    f"<https://www.run.house/docs/api/python/cluster#runhouse.ondemand_cluster>`__)."
                )
            raise e

        self._update_from_sky_status()

        if self.domain:
            logger.info(
                f"Cluster has been launched with the custom domain '{self.domain}'. "
                "Please add an A record to your DNS provider to point this domain to the cluster's "
                f"public IP address ({self.address}) to ensure successful requests."
            )

        self.restart_server()

        if rns_client.autosave_resources():
            self.save()

        return self

    def keep_warm(self, mins: int = -1):
        """Keep the cluster warm for given number of minutes after inactivity.

        Args:
            mins (int): Amount of time (in min) to keep the cluster warm after inactivity.
                If set to -1, keep cluster warm indefinitely. (Default: `-1`)
        """
        self.autostop_mins = mins
        return self

    def teardown(self):
        """Teardown cluster.

        Example:
            >>> rh.ondemand_cluster("rh-cpu").teardown()
        """
        try:
            cluster_status_data = self.status()
            status_data = {
                "status": ResourceServerStatus.terminated,
                "resource_type": self.__class__.__base__.__name__.lower(),
                "data": cluster_status_data,
            }
            cluster_uri = rns_client.format_rns_address(self.rns_address)
            api_server_url = rns_client.api_server_url
            status_resp = requests.post(
                f"{api_server_url}/resource/{cluster_uri}/cluster/status",
                data=json.dumps(status_data),
                headers=rns_client.request_headers(),
            )
            # 404 means that the cluster is not saved in den, it is fine that the status is not updated.
            if status_resp.status_code not in [200, 404]:
                logger.warning("Failed to update Den with terminated cluster status")
        except Exception as e:
            logger.warning(e)

        # Stream logs
        sky.down(self.name)
        self.address = None

    def teardown_and_delete(self):
        """Teardown cluster and delete it from configs.

        Example:
            >>> rh.ondemand_cluster("rh-cpu").teardown_and_delete()
        """
        self.teardown()
        rns_client.delete_configs(resource=self)

    @contextlib.contextmanager
    def pause_autostop(self):
        """Context manager to temporarily pause autostop.

        Example:
            >>> with rh.ondemand_cluster.pause_autostop():
            >>>     rh.ondemand_cluster.run(["python train.py"])
        """
        sky.autostop(self.name, idle_minutes=-1)
        yield
        sky.autostop(self.name, idle_minutes=self._autostop_mins, down=True)

    # ----------------- SSH Methods ----------------- #

    @staticmethod
    def cluster_ssh_key(path_to_file: Path):
        """Retrieve SSH key for the cluster.

        Args:
            path_to_file (Path): Path of the private key associated with the cluster.

        Example:
            >>> ssh_priv_key = rh.ondemand_cluster("rh-cpu").cluster_ssh_key("~/.ssh/id_rsa")
        """
        try:
            f = open(path_to_file, "r")
            private_key = f.read()
            return private_key
        except FileNotFoundError:
            raise Exception(f"File with ssh key not found in: {path_to_file}")

    def ssh(self, node: str = None):
        """SSH into the cluster.

        Args:
            node: Node to SSH into. If no node is specified, will SSH onto the head node.
                (Default: ``None``)

        Example:
            >>> rh.ondemand_cluster("rh-cpu").ssh()
            >>> rh.ondemand_cluster("rh-cpu", node="3.89.174.234").ssh()
        """
        if self.provider == "kubernetes":
            command = f"kubectl get pods | grep {self.name}"

            try:

                output = subprocess.check_output(command, shell=True, text=True)

                lines = output.strip().split("\n")
                if lines:
                    pod_name = lines[0].split()[0]
                else:
                    logger.info("No matching pods found.")
            except subprocess.CalledProcessError as e:
                raise Exception(f"Error: {e}")

            cmd = f"kubectl exec -it {pod_name} -- /bin/bash"
            subprocess.run(cmd, shell=True, check=True)

        else:
            # If SSHing onto a specific node, which requires the default sky public key for verification
            from runhouse.resources.hardware.sky_command_runner import SshMode

            sky_key = Path(
                self.creds_values.get("ssh_private_key", self.DEFAULT_KEYFILE)
            ).expanduser()

            if not sky_key.exists():
                raise FileNotFoundError(f"Expected default sky key in path: {sky_key}")

            runner = self._command_runner(node=node)
            cmd = runner.run(
                cmd="bash --rcfile <(echo '. ~/.bashrc; conda deactivate')",
                ssh_mode=SshMode.INTERACTIVE,
                port_forward=None,
                return_cmd=True,
            )
            subprocess.run(cmd, shell=True)

    def _ping(self, timeout=5, retry=False):
        if super()._ping(timeout=timeout, retry=False):
            return True

        if retry:
            self._update_from_sky_status(dryrun=False)
            return super()._ping(timeout=timeout, retry=False)
        return False



import pprint
import sys
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from runhouse.globals import obj_store, rns_client
from runhouse.logger import get_logger

from runhouse.rns.top_level_rns_fns import (
    resolve_rns_path,
    save,
    split_rns_name_and_path,
)
from runhouse.rns.utils.api import (
    load_resp_content,
    read_resp_data,
    ResourceAccess,
    ResourceVisibility,
)

logger = get_logger(__name__)


class Resource:
    RESOURCE_TYPE = "resource"

    def __init__(
        self,
        name: Optional[str] = None,
        dryrun: bool = False,
        access_level: ResourceAccess = ResourceAccess.WRITE,
        visibility: ResourceVisibility = ResourceVisibility.PRIVATE,
        **kwargs,
    ):
        """
        Runhouse abstraction for objects that can be saved, shared, and reused.

        Args:
            name (Optional[str], optional): Name to assign the resource. (Default: None)
            dryrun (bool, optional): Whether to create the resource object, or load the object as a dryrun.
                (Default: ``False``)
            access_level (:obj:`ResourceAccess`, optional): Access level to provide for the resource.
                (Default: ``ResourceAccess.WRITE``)
            visibility (:obj:`ResourceVisibility`, optional): Type of visibility to provide for the resource.
                (Default: ``ResourceVisibility.PRIVATE``)
        """
        self._name, self._rns_folder = None, None
        if name is not None:
            # TODO validate that name complies with a simple regex
            if name.startswith("/builtins/"):
                name = name[len("/builtins/") :]
            if name[0] == "^" and name != "^":
                name = name[1:]
            self._name, self._rns_folder = rns_client.split_rns_name_and_path(
                rns_client.resolve_rns_path(name)
            )

        self.dryrun = dryrun
        self.access_level = access_level
        self._visibility = visibility

    # TODO add a utility to allow a parameter to be specified as "default" and then use the default value

    @property
    def config_for_rns(self):
        # Added for BC for version 0.0.20
        return self.config(condensed=False)

    def config(self, condensed=True):
        config = {
            "name": self.rns_address or self.name,
            "resource_type": self.RESOURCE_TYPE,
            "resource_subtype": self.__class__.__name__,
        }
        self.save_attrs_to_config(
            config,
            [
                "visibility",  # Handles Enum to string conversion
            ],
        )
        return config

    def _resource_string_for_subconfig(
        self, resource: Union[None, str, "Resource"], condensed=True
    ):
        """Returns a string representation of a sub-resource for use in a config."""
        if resource is None or isinstance(resource, str):
            return resource
        if isinstance(resource, Resource):
            if condensed and resource.rns_address:
                # We operate on the assumption that rns_address is only populated once a resource has been saved.
                # That way, if rns_address is not None, we have reasonable likelihood that the resource was saved and
                # we can just pass the address. The only exception here is if the resource is a built-in.
                if resource.rns_address.startswith("^"):
                    # Fork the resource if it's a built-in and consider it a new resource
                    resource._rns_folder = None
                    return resource.config(condensed)
                return resource.rns_address
            else:
                # If the resource doesn't have an rns_address, we consider it unsaved and put the whole config into
                # the parent config.
                return resource.config(condensed)
        raise ValueError(
            f"Resource {resource} is not a valid sub-resource for {self.__class__.__name__}"
        )

    @property
    def rns_address(self):
        """Traverse up the filesystem until reaching one of the directories in rns_base_folders,
        then compute the relative path to that."""
        if (
            self.name is None or self._rns_folder is None
        ):  # Anonymous folders have no rns address
            return None

        return str(Path(self._rns_folder) / self.name)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        # Split the name and rns path if path is given (concat with current_folder if just stem is given)
        if name is None:
            self._name = None
        else:
            self._name, self._rns_folder = split_rns_name_and_path(
                resolve_rns_path(name)
            )

    @property
    def visibility(self):
        return self._visibility

    @visibility.setter
    def visibility(self, visibility):
        self._visibility = visibility

    @rns_address.setter
    def rns_address(self, new_address):
        self.name = new_address  # Note, this saves the resource to the new address!

    def _save_sub_resources(self, folder: str = None):
        """Overload by child resources to save any resources they hold internally."""
        pass

    def pin(self):
        """Write the resource to the object store."""
        from runhouse.resources.hardware.utils import _current_cluster

        if _current_cluster():
            if obj_store.has_local_storage:
                obj_store.put_local(self._name, self)
            else:
                obj_store.put(self._name, self)
        else:
            raise ValueError("Cannot pin a resource outside of a cluster.")

    def refresh(self):
        """Update the resource in the object store."""
        from runhouse.resources.hardware.utils import _current_cluster

        if _current_cluster():
            return obj_store.get(self._name)
        else:
            return self

    def save(self, name: str = None, overwrite: bool = True, folder: str = None):
        """Register the resource, saving it to local working_dir config and RNS config store. Uses the resource's
        `self.config()` to generate the dict to save."""

        # add this resource this run's downstream artifact registry if it's being saved as part of a run
        rns_client.add_downstream_resource(name or self.name)

        self._save_sub_resources(folder)
        if name:
            self.name = name

        # TODO handle self.access == 'read' instead of this weird overwrite argument
        save(self, overwrite=overwrite, folder=folder)

        return self

    def __str__(self):
        return pprint.pformat(self.config())

    @classmethod
    def _check_for_child_configs(cls, config: dict):
        """Overload by child resources to load any resources they hold internally."""
        return config

    @classmethod
    def _compare_config_with_alt_options(cls, config, alt_options):
        """Overload by child resources to compare their config with the alt_options. If the user specifies alternate
        options, compare the config with the options. It's generally up to the child class to decide how to handle the
        options, but default behavior is provided. The default behavior simply checks if any of the alt_options are
        present in the config (with awareness of resources), and if their values differ, return None.

        If the child class returns None, it's deciding to override the config
        with the options. If the child class returns a config, it's deciding to use the config and ignore the options
        (or somehow incorporate them, rarely). Note that if alt_options are provided and the config is not found,
        no error is raised, while if alt_options are not provided and the config is not found, an error is raised.
        """

        def str_dict_or_resource_to_str(val):
            if isinstance(val, Resource):
                return val.rns_address
            elif isinstance(val, dict):
                # This can either be a sub-resource which hasn't been converted to a resource yet, or an
                # actual user-provided dict
                if "rns_address" in val:
                    return val["rns_address"]
                if "name" in val:
                    # convert a user-provided name to an rns_address
                    return rns_client.resolve_rns_path(val["name"])
                else:
                    return val
            elif isinstance(val, list):
                val = [str(item) if isinstance(item, int) else item for item in val]
            return val

        for key, value in alt_options.items():
            if key in config:
                if str_dict_or_resource_to_str(value) != str_dict_or_resource_to_str(
                    config[key]
                ):
                    return None
            else:
                return None
        return config

    @classmethod
    def from_name(
        cls,
        name: str,
        load_from_den: bool = True,
        dryrun: bool = False,
        _alt_options: Dict = None,
        _resolve_children: bool = True,
    ):
        """Load existing Resource via its name.

        Args:
            name (str): Name of the resource to load from name.
            load_from_den (bool, optional): Whether to try loading the module from Den. (Default: ``True``)
            dryrun (bool, optional): Whether to construct the object or load as dryrun. (Default: ``False``)
        """
        # TODO is this the right priority order?
        from runhouse.resources.hardware.utils import _current_cluster

        if _current_cluster() and obj_store.contains(name):
            return obj_store.get(name)

        config = rns_client.load_config(name=name, load_from_den=load_from_den)

        if _alt_options:
            config = cls._compare_config_with_alt_options(config, _alt_options)
            if not config:
                return None
        if not config:
            raise ValueError(f"Resource {name} not found.")
        config["name"] = name

        if _resolve_children:
            config = cls._check_for_child_configs(config)

        # Add this resource's name to the resource artifact registry if part of a run
        rns_client.add_upstream_resource(name)

        # Uses child class's from_config
        return cls.from_config(
            config=config, dryrun=dryrun, _resolve_children=_resolve_children
        )

    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        """Load or construct resource from config.

        Args:
            config (Dict): Resource config.
            dryrun (bool, optional): Whether to construct resource or load as dryrun (Default: ``False``)
        """
        resource_type = config.pop("resource_type", None)
        dryrun = config.pop("dryrun", False) or dryrun

        if resource_type == "resource":
            return Resource(**config, dryrun=dryrun)

        resource_class = getattr(
            sys.modules["runhouse"], resource_type.capitalize(), None
        )
        if not resource_class:
            raise TypeError(f"Could not find module associated with {resource_type}")

        if _resolve_children:
            config = resource_class._check_for_child_configs(config)

        loaded = resource_class.from_config(
            config=config,
            dryrun=dryrun,
            _resolve_children=_resolve_children,
        )
        if loaded.name:
            rns_client.add_upstream_resource(loaded.name)
        return loaded

    def unname(self):
        """Remove the name of the resource. This changes the resource name to anonymous and deletes any local
        or RNS configs for the resource."""
        self.delete_configs()
        self._name = None

    def history(self, limit: int = None) -> List[Dict]:
        """Return the history of the resource, including specific config fields (e.g. folder path) and which runs
        have overwritten it.

        Args:
            limit (int, optional): If specified, return the last ``limit`` number of entries in the history.
                Otherwise, return the entire history. (Default: ``None``)
        """
        if not self.rns_address:
            raise ValueError("Resource must have a name in order to have a history")

        if self.rns_address[:2] == "~/":
            raise ValueError(
                "Resource must be saved to Den (not local) in order to have a history"
            )

        resource_uri = rns_client.resource_uri(self.rns_address)
        base_uri = f"{rns_client.api_server_url}/resource/history/{resource_uri}"
        uri = f"{base_uri}?limit={limit}" if limit else base_uri

        resp = rns_client.session.get(uri, headers=rns_client.request_headers())
        if resp.status_code != 200:
            logger.warning(
                f"Received [{resp.status_code}] from Den GET '{uri}': No resource history found: {load_resp_content(resp)}"
            )
            return []

        resource_history = read_resp_data(resp)
        return resource_history

    # TODO delete sub-resources
    def delete_configs(self):
        """Delete the resource's config from local working_dir and RNS config store."""
        rns_client.delete_configs(resource=self)

    def save_attrs_to_config(self, config: Dict, attrs: List[str]):
        """Save the given attributes to the config"""
        for attr in attrs:
            val = self.__getattribute__(attr)
            if val is not None:
                if isinstance(val, Enum):
                    val = val.value
                config[attr] = val

    def is_local(self):
        return (
            hasattr(self, "install_target")
            and isinstance(self.install_target, str)
            and self.install_target.startswith("~")
            or hasattr(self, "system")
            and self.system == "file"
        )

    # TODO [DG] Implement proper sharing of subresources (with an overload of some kind)
    def share(
        self,
        users: Union[str, List[str]] = None,
        access_level: Union[ResourceAccess, str] = ResourceAccess.READ,
        visibility: Optional[Union[ResourceVisibility, str]] = None,
        notify_users: bool = True,
        headers: Optional[Dict] = None,
    ) -> Tuple[Dict[str, ResourceAccess], Dict[str, ResourceAccess]]:
        """Grant access to the resource for a list of users (or a single user). By default, the user will
        receive an email notification of access (if they have a Runhouse account) or instructions on creating
        an account to access the resource. If ``visibility`` is set to ``public``, users will not be notified.

        .. note::
            You can only grant access to other users if you have write access to the resource.

        Args:
            users (Union[str, list], optional): Single user or list of user emails and / or runhouse account usernames.
                If none are provided and ``visibility`` is set to ``public``, resource will be made publicly
                available to all users. (Default: ``None``)
            access_level (:obj:`ResourceAccess`, optional): Access level to provide for the resource.
                (Default: ``read``).
            visibility (:obj:`ResourceVisibility`, optional): Type of visibility to provide for the shared
                resource. By default, the visibility is private. (Default: ``None``)
            notify_users (bool, optional): Whether to send an email notification to users who have been given access.
                Note: This is relevant for resources which are not ``shareable``. (Default: ``True``)
            headers (Dict, optional): Request headers to provide for the request to RNS. Contains the user's auth token.
                Example: ``{"Authorization": f"Bearer {token}"}``

        Returns:
            Tuple(Dict, Dict, Set):

            `added_users`:
                Users who already have a Runhouse account and have been granted access to the resource.
            `new_users`:
                Users who do not have Runhouse accounts and received notifications via their emails.
            `valid_users`:
                Set of valid usernames and emails from ``users`` parameter.

        Example:
            >>> # Write access to the resource for these specific users.
            >>> # Visibility will be set to private (users can search for and view resource in Den dashboard)
            >>> my_resource.share(users=["username1", "user2@gmail.com"], access_level='write')

            >>> # Make resource public, with read access to the resource for all users
            >>> my_resource.share(visibility='public')
        """
        if self.name is None:
            raise ValueError("Resource must have a name in order to share")

        if users is None and visibility is None:
            raise ValueError(
                "Must specify `visibility` for the resource if no users are provided."
            )

        if hasattr(self, "system") and self.system in ["ssh", "sftp"]:
            logger.warning(
                "Sharing a resource located on a cluster is not recommended. For persistence, we suggest"
                "saving to a cloud storage system (ex: `s3` or `gs`). You can copy your cluster based "
                f"{self.RESOURCE_TYPE} to your desired storage provider using the `.to()` method. "
                f"For example: `{self.RESOURCE_TYPE}.to(system='rh-cpu')`"
            )

        if self.is_local():
            if self.RESOURCE_TYPE == "package":
                raise TypeError(
                    f"Unable to share a local {self.RESOURCE_TYPE}. Please make sure the {self.RESOURCE_TYPE} is "
                    f"located on a cluster. You can use the `.to()` method to do so. "
                    f"For example: `{self.name}.to(system='rh-cpu')`"
                )
            else:
                raise TypeError(
                    f"Unable to share a local {self.RESOURCE_TYPE}. Please make sure the {self.RESOURCE_TYPE} is "
                    f"located on a cluster or a remote system. You can use the `.to()` method to do so. "
                    f"For example: `{self.name}.to(system='s3')`"
                )

        if isinstance(access_level, str):
            access_level = ResourceAccess(access_level)

        if visibility is not None:
            # Update the resource in Den with this global visibility value
            self.visibility = visibility

            logger.debug(f"Updating resource with visibility: {self.visibility}")

        self.save()

        if isinstance(users, str):
            users = [users]

        added_users, new_users, valid_users = rns_client.grant_resource_access(
            rns_address=self.rns_address,
            user_emails=users,
            access_level=access_level,
            notify_users=notify_users,
            headers=headers,
        )
        return added_users, new_users, valid_users

    def revoke(
        self, users: Union[str, List[str]] = None, headers: Optional[Dict] = None
    ):
        """Revoke access to the resource.

        Args:
            users (Union[str, str], optional): List of user emails and / or runhouse account usernames
                (or a single user). If no users are specified will revoke access for all users. (Default: ``None``)
            headers (Optional[Dict]): Request headers to provide for the request to RNS. Contains the user's auth token.
                Example: ``{"Authorization": f"Bearer {token}"}``
        """
        if isinstance(users, str):
            users = [users]

        request_uri = rns_client.resource_uri(self.rns_address)
        uri = f"{rns_client.api_server_url}/resource/{request_uri}/users/access"
        resp = rns_client.session.put(
            uri,
            json={"users": users, "access_level": ResourceAccess.DENIED},
            headers=headers or rns_client.request_headers(),
        )

        if resp.status_code != 200:
            raise Exception(
                f"Received [{resp.status_code}] from Den PUT '{uri}': Failed to revoke access for resource: {load_resp_content(resp)}"
            )



import os

from pathlib import Path
from typing import Dict

import pytest

import runhouse as rh

import tests.test_resources.test_resource

from tests.conftest import init_args


def _get_env_var_value(env_var):
    import os

    return os.environ[env_var]


def _uninstall_env(env, cluster):
    for req in env.reqs:
        if "/" not in req:
            cluster.run([f"pip uninstall {req} -y"], env=env)


def np_summer(a, b):
    import numpy as np

    return int(np.sum([a, b]))


def scipy_import():
    import scipy

    return str(scipy.__version__)


@pytest.mark.envtest
class TestEnv(tests.test_resources.test_resource.TestResource):
    MAP_FIXTURES = {"resource": "env"}

    UNIT = {
        "env": [
            "unnamed_env",
            "named_env",
            "base_conda_env",
            "named_conda_env_from_dict",
            "conda_env_from_local",
            "conda_env_from_path",
        ]
    }
    LOCAL = {
        "env": ["unnamed_env", "named_env"],
        "cluster": [
            "docker_cluster_pk_ssh_no_auth",
        ]
        # TODO: extend envs to "base_conda_env", "named_conda_env_from_dict"],
        # and add local clusters once conda docker container is set up
    }
    MINIMAL = {
        "env": [
            "unnamed_env",
            "named_env",
            "base_conda_env",
            "named_conda_env_from_dict",
        ],
        "cluster": ["ondemand_aws_docker_cluster"],
    }
    RELEASE = {
        "env": [
            "unnamed_env",
            "named_env",
            "base_conda_env",
            "named_conda_env_from_dict",
        ],
        "cluster": [
            "ondemand_aws_docker_cluster",
            "static_cpu_pwd_cluster",
        ],
    }
    MAXIMAL = {
        "env": [
            "unnamed_env",
            "named_env",
            "base_conda_env",
            "named_conda_env_from_dict",
            "conda_env_from_local",
            "conda_env_from_path",
        ],
        "cluster": [
            "ondemand_aws_docker_cluster",
            "ondemand_gcp_cluster",
            "ondemand_k8s_cluster",
            "ondemand_k8s_docker_cluster",
            "ondemand_aws_https_cluster_with_auth",
            "static_cpu_pwd_cluster",
            "multinode_cpu_docker_conda_cluster",
            "docker_cluster_pk_ssh_no_auth",
            "docker_cluster_pwd_ssh_no_auth",
            "docker_cluster_pk_ssh_den_auth",
        ],
    }

    @pytest.mark.level("unit")
    def test_env_factory_and_properties(self, env):
        assert isinstance(env, rh.Env)

        args = init_args[id(env)]
        if "reqs" in args:
            assert set(args["reqs"]).issubset(env.reqs)

        if isinstance(env, rh.CondaEnv):
            assert env.conda_yaml
            assert isinstance(env.conda_yaml, Dict)
            assert set(["dependencies", "name"]).issubset(set(env.conda_yaml.keys()))

        if "name" not in args:
            assert not env.name

    @pytest.mark.level("unit")
    def test_env_conda_env_factories(self):
        name = "env_name"
        conda_env = {
            "name": "conda_env_name",
            "reqs": "pytest",
        }

        env = rh.env(name=name, conda_env=conda_env)
        conda_env = rh.conda_env(name=name, conda_env=conda_env)

        assert env.config() == conda_env.config()

    @pytest.mark.level("local")
    def test_env_to_cluster(self, env, cluster):
        env.to(cluster, force_install=True)

        for req in env.reqs:
            if not req == "./":
                res = cluster.run([f"pip freeze | grep {req}"], env=env)
                assert res[0][0] == 0  # installed properly

        _uninstall_env(env, cluster)

    @pytest.mark.skip("Running into s3 folder issue")
    @pytest.mark.level("minimal")
    def test_env_to_fs_to_cluster(self, env, cluster):
        s3_env = env.to("s3", force_install=True)
        for req in s3_env.reqs:
            if isinstance(req, rh.Package) and isinstance(
                req.install_target, rh.Folder
            ):
                assert req.install_target.system == "s3"
                assert req.install_target.exists_in_system()
                count += 1
        assert count >= 1

        folder_name = "test_package"
        count = 0
        conda_env_cluster = s3_env.to(
            system=cluster, path=folder_name, force_install=True
        )
        for req in conda_env_cluster.reqs:
            if isinstance(req, rh.Package) and isinstance(
                req.install_target, rh.Folder
            ):
                assert req.install_target.system == cluster
                count += 1
        assert count >= 1

        assert "sample_file_0.txt" in cluster.run([f"ls {folder_name}"])[0][1]
        cluster.run([f"rm -r {folder_name}"])

    @pytest.mark.level("local")
    def test_addtl_env_reqs(self, env, cluster):
        package = "jedi"
        env.reqs = env.reqs + [package] if env.reqs else [package]
        env.to(cluster, force_install=True)

        res = cluster.run([f"pip freeze | grep {package}"], env=env)
        assert res[0][0] == 0
        assert package in env.reqs

        _uninstall_env(env, cluster)

    @pytest.mark.level("local")
    def test_fn_to_env(self, env, cluster):
        package = "numpy"
        env.reqs = env.reqs + [package] or [package]
        fn = rh.function(np_summer).to(system=cluster, env=env, force_install=True)
        assert fn(1, 4) == 5

    @pytest.mark.level("local")
    def test_env_vars_dict(self, env, cluster):
        test_env_var = "TEST_ENV_VAR"
        test_value = "value"
        env.env_vars = {test_env_var: test_value}

        get_env_var_cpu = rh.function(_get_env_var_value).to(
            system=cluster, env=env, force_install=True
        )
        res = get_env_var_cpu(test_env_var)

        assert res == test_value

        _uninstall_env(env, cluster)

    @pytest.mark.level("local")
    def test_env_vars_file(self, env, cluster, tmp_path):
        env_file = str(tmp_path / ".env")
        contents = [
            "# comment",
            "",
            "ENV_VAR1=value",
            "# comment with =",
            "ENV_VAR2 =val2",
        ]
        with open(env_file, "w") as f:
            for line in contents:
                f.write(line + "\n")

        env.env_vars = env_file

        get_env_var_cpu = rh.function(_get_env_var_value).to(
            system=cluster, env=env, force_install=True
        )
        assert get_env_var_cpu("ENV_VAR1") == "value"
        assert get_env_var_cpu("ENV_VAR2") == "val2"

        os.remove(env_file)
        assert not Path(env_file).exists()

        _uninstall_env(env, cluster)

    @pytest.mark.level("local")
    def test_working_dir_env(self, env, cluster, tmp_path):
        dir_name = "test_working_dir"
        working_dir = tmp_path / dir_name
        working_dir.mkdir(exist_ok=True)
        env.working_dir = str(working_dir)

        assert str(working_dir) in env.reqs

        # Send the env to the cluster, save the dir in the main working directory (~) of the cluster
        env.to(cluster, path=dir_name, force_install=True)
        assert working_dir.name in cluster.run(["ls"])[0][1]

        cluster.run([f"rm -r {working_dir.name}"])
        assert working_dir.name not in cluster.run(["ls"])[0][1]

        _uninstall_env(env, cluster)

    @pytest.mark.level("local")
    def test_secrets_env(self, env, cluster):
        path_secret = rh.provider_secret(
            "lambda", values={"api_key": "test_api_key"}
        ).write(path="~/lambda_keys")
        api_key_secret = rh.provider_secret(
            "openai", values={"api_key": "test_openai_key"}
        )
        named_secret = (
            rh.provider_secret("huggingface", values={"token": "test_hf_token"})
            .write(path="~/hf_token")
            .save()
        )
        secrets = [path_secret, api_key_secret, named_secret.provider]

        env.secrets = secrets
        get_env_var_cpu = rh.function(_get_env_var_value).to(
            system=cluster, env=env, force_install=True
        )

        for secret in secrets:
            name = (
                secret if isinstance(secret, str) else (secret.name or secret.provider)
            )
            assert cluster.get(name)

            if isinstance(secret, str):
                secret = rh.Secret.from_name(secret)

            if secret.path:
                assert rh.folder(path=secret.path, system=cluster).exists_in_system()
            else:
                env_vars = secret.env_vars or secret._DEFAULT_ENV_VARS
                for _, var in env_vars.items():
                    assert get_env_var_cpu(var)

        named_secret.delete()

    @pytest.mark.level("local")
    def test_env_run_cmd(self, env, cluster):
        test_env_var = "ENV_VAR"
        test_value = "env_val"
        env.env_vars = {test_env_var: test_value}

        env.to(cluster)
        res = cluster.run(["echo $ENV_VAR"], env=env)

        assert res[0][0] == 0  # returncode
        assert "env_val" in res[0][1]  # stdout

    @pytest.mark.level("local")
    @pytest.mark.parametrize(
        "cmd",
        [
            "export ENV_VAR=env_val",
            "pip install numpy; echo $ENV_VAR",
            "pip freeze | grep numpy",
        ],
    )
    def test_env_run_shell_cmds(self, env, cluster, cmd):
        env.to(cluster)

        res = cluster.run([cmd], env=env)
        assert res[0][0] == 0

    @pytest.mark.level("local")
    def test_env_to_with_provider_secret(self, cluster):
        os.environ["HF_TOKEN"] = "test_hf_token"
        env = rh.env(name="hf_env", secrets=["huggingface"])
        env.to(cluster)

    @pytest.mark.level("local")
    def test_env_in_function_factory(self, cluster):
        remote_function = rh.function(scipy_import, env=["scipy"]).to(system=cluster)
        assert remote_function() is not None



import os

import pytest

import runhouse as rh

from tests.conftest import init_args


def _get_conda_env(name="rh-test", python_version="3.10.9"):
    conda_env = {
        "name": name,
        "channels": ["defaults"],
        "dependencies": [
            f"python={python_version}",
        ],
    }
    return conda_env


@pytest.fixture(scope="function")
def env(request):
    """Parametrize over multiple envs - useful for running the same test on multiple envs."""
    return request.getfixturevalue(request.param)


@pytest.fixture(scope="function")
def unnamed_env():
    args = {"reqs": ["npm"]}
    env = rh.env(**args)
    init_args[id(env)] = args
    return env


@pytest.fixture(scope="function")
def named_env():
    args = {"reqs": ["npm"], "name": "named_env"}
    env = rh.env(**args)
    init_args[id(env)] = args
    return env


@pytest.fixture(scope="function")
def base_conda_env():
    args = {"name": "conda_base", "reqs": ["pytest", "npm"]}
    env = rh.conda_env(**args)
    init_args[id(env)] = args
    return env


@pytest.fixture(scope="function")
def named_conda_env_from_dict():
    env_name = "conda_from_dict"
    conda_dict = _get_conda_env(name=f"{env_name}_env")

    args = {"name": env_name, "conda_env": conda_dict}
    env = rh.conda_env(**args)
    init_args[id(env)] = args
    return env


@pytest.fixture(scope="function")
def conda_env_from_path():
    env_name = "conda_from_path"
    file_path = os.path.join(os.path.dirname(__file__), "assets", "test_conda_env.yml")

    args = {"name": env_name, "conda_env": file_path}
    env = rh.conda_env(**args)
    init_args[id(env)] = args
    yield env


@pytest.fixture(scope="session")
def _local_conda_env():
    env_name = "test_conda_local_env"
    try:
        os.system(f"conda create -n {env_name} -y python==3.10.9")
        yield
    finally:
        os.system(f"conda env remove -n {env_name} -y")


@pytest.fixture(scope="function")
def conda_env_from_local(_local_conda_env):
    env_name = "test_conda_local_env"

    args = {"name": env_name, "conda_env": env_name}
    env = rh.conda_env(**args)
    init_args[id(env)] = args
    return env



import datetime
import importlib
import logging
import math
import shlex
import subprocess
import time
import webbrowser
from pathlib import Path
from typing import Any, Dict, List, Optional

import ray

import requests
import rich.markdown

import typer
import yaml
from rich.console import Console

import runhouse as rh

import runhouse.rns.login

from runhouse import __version__, cluster, Cluster, configs
from runhouse.constants import (
    BULLET_UNICODE,
    DOUBLE_SPACE_UNICODE,
    RAY_KILL_CMD,
    RAY_START_CMD,
    SERVER_LOGFILE,
    SERVER_START_CMD,
    SERVER_STOP_CMD,
    START_NOHUP_CMD,
    START_SCREEN_CMD,
)
from runhouse.globals import rns_client
from runhouse.logger import get_logger
from runhouse.resources.hardware.ray_utils import (
    check_for_existing_ray_instance,
    kill_actors,
)

# create an explicit Typer application
app = typer.Typer(add_completion=False)

# For printing with typer
console = Console()

logger = get_logger(__name__)


@app.command()
def login(
    token: Optional[str] = typer.Argument(None, help="Your Runhouse API token"),
    sync_secrets: Optional[bool] = typer.Option(
        False,
        "--sync-secrets",
        help="Whether to sync secrets. You will be prompted whether to upload local secrets or download saved secrets",
    ),
    yes: Optional[bool] = typer.Option(
        False, "--yes", "-y", help="Sets any confirmations to 'yes' automatically."
    ),
):
    """Login to Runhouse. Validates token provided, with options to upload or download stored secrets or config between
    local environment and Runhouse / Vault.
    """
    valid_token: str = (
        runhouse.rns.login.login(
            token=token,
            download_config=True,
            upload_config=True,
            download_secrets=True,
            upload_secrets=True,
            from_cli=True,
        )
        if yes
        else runhouse.rns.login.login(
            token=token,
            interactive=True,
            ret_token=True,
            from_cli=True,
            sync_secrets=sync_secrets,
        )
    )

    if valid_token:
        webbrowser.open(f"{configs.get('dashboard_url')}/dashboard?token={valid_token}")
        raise typer.Exit()
    else:
        raise typer.Exit(code=1)


@app.command()
def logout():
    """Logout of Runhouse. Provides options to delete locally configured secrets and local Runhouse configs"""
    runhouse.rns.login.logout(interactive=True)
    raise typer.Exit()


@app.command()
def notebook(
    cluster_name: str, up: bool = typer.Option(False, help="Start the cluster")
):
    """Open a Jupyter notebook on a cluster."""
    c = cluster(name=cluster_name)
    if up:
        c.up_if_not()
    if not c.is_up():
        console.print(
            f"Cluster {cluster_name} is not up. Please run `runhouse notebook {cluster_name} --up` to bring "
            f"it up if it is an on-demand cluster."
        )
        raise typer.Exit(1)
    c.notebook()


@app.command()
def ssh(cluster_name: str, up: bool = typer.Option(False, help="Start the cluster")):
    """SSH into a cluster created elsewhere (so `ssh cluster` doesn't work out of the box) or not yet up."""
    try:
        c = cluster(name=cluster_name)

        if not c.is_shared:
            if up:
                try:
                    c.up_if_not()
                except NotImplementedError:
                    console.print(
                        f"Cluster {cluster_name} is not an on-demand cluster, so it can't be brought up automatically."
                        f"Please start it manually and re-save the cluster with the new connection info in Python."
                    )
                    raise typer.Exit(1)
            elif not c.is_up():
                console.print(
                    f"Cluster {cluster_name} is not up. Please run `runhouse ssh {cluster_name} --up`."
                )
                raise typer.Exit(1)

        c.ssh()
    except ValueError:
        import sky

        from runhouse import OnDemandCluster
        from runhouse.constants import DEFAULT_SSH_PORT
        from runhouse.resources.hardware.utils import _run_ssh_command

        state = sky.status(cluster_names=[cluster_name], refresh=False)

        if len(state) == 0:
            console.print(
                f"Could not load cluster called {cluster_name}. Cluster must either be saved to Den, "
                "or be an ondemand cluster that is currently up."
            )

            raise typer.Exit(1)

        resource_handle = state[0].get("handle", {})

        sky_ssh_config = Path(f"~/.sky/generated/{cluster_name}.yml").expanduser()

        if sky_ssh_config.exists():
            with open(sky_ssh_config, "r", encoding="utf-8") as f:
                ssh_config = yaml.safe_load(f)
            ssh_proxy_command = ssh_config["auth"].get("ssh_proxy_command", None)
            if ssh_proxy_command:
                ssh_proxy_command = ssh_proxy_command.replace(
                    "skypilot:ssh_user", resource_handle.ssh_user
                )

        else:
            ssh_proxy_command = None

        _run_ssh_command(
            address=resource_handle.head_ip,
            ssh_user=resource_handle.ssh_user or "ubuntu",
            ssh_port=resource_handle.stable_ssh_ports[0] or DEFAULT_SSH_PORT,
            ssh_private_key=OnDemandCluster.DEFAULT_KEYFILE,
            docker_user=resource_handle.docker_user,
            ssh_proxy_command=ssh_proxy_command,
        )


###############################
# Status helping functions
###############################


def _adjust_resource_type(resource_type: str):
    """
    status helping function. transforms a str form runhouse.resources.{X.Y...}.resource_type to runhouse.resource_type
    """
    try:
        resource_type = resource_type.split(".")[-1]
        getattr(importlib.import_module("runhouse"), resource_type)
        return f"runhouse.{resource_type}"
    except AttributeError:
        return resource_type


def _resource_name_to_rns(name: str):
    """
    If possible, transform the resource name to a rns address.
    If not, return the name as is (it is the key in the object store).
    """
    resource_config = rns_client.load_config(name)
    if resource_config and resource_config.get("resource_type") != "env":
        return resource_config.get("name")
    else:
        return name


def _print_cluster_config(cluster_config: Dict):
    """
    Helping function to the `_print_status` which prints the relevant info from the cluster config.
    """
    # TODO [SB]: need to modify printing format (colour palette etc).

    top_level_config = [
        "server_port",
        "den_auth",
        "server_connection_type",
    ]

    backend_config = ["resource_subtype", "domain", "server_host", "ips"]

    if cluster_config.get("resource_subtype") != "Cluster":
        backend_config.append("autostop_mins")

    if cluster_config.get("default_env") and isinstance(
        cluster_config.get("default_env"), Dict
    ):
        cluster_config["default_env"] = cluster_config["default_env"]["name"]

    for key in top_level_config:
        console.print(
            f"{BULLET_UNICODE} {key.replace('_', ' ')}: {cluster_config[key]}"
        )

    console.print(f"{BULLET_UNICODE} backend config:")
    for key in backend_config:
        if key == "autostop_mins" and cluster_config[key] == -1:
            console.print(
                f"{DOUBLE_SPACE_UNICODE}{BULLET_UNICODE} {key.replace('_', ' ')}: autostop disabled"
            )
        else:
            console.print(
                f"{DOUBLE_SPACE_UNICODE}{BULLET_UNICODE} {key.replace('_', ' ')}: {cluster_config[key]}"
            )


def _print_envs_info(
    env_servlet_processes: Dict[str, Dict[str, Any]], current_cluster: Cluster
):
    """
    Prints info about the envs in the current_cluster: resources in each env, the CPU usage and GPU usage of the env
    (if exists)
    """

    # Print headline
    envs_in_cluster_headline = "Serving 🍦 :"
    console.print(envs_in_cluster_headline)

    env_resource_mapping = {
        env: env_servlet_processes[env]["env_resource_mapping"]
        for env in env_servlet_processes
    }

    if len(env_resource_mapping) == 0:
        console.print("This cluster has no environment nor resources.")

    first_envs_to_print = []

    # First: if the default env does not have resources, print it.
    default_env_name = current_cluster.default_env.name
    if len(env_resource_mapping[default_env_name]) <= 1:
        # case where the default env doesn't hve any other resources, apart from the default env itself.
        console.print(f"{BULLET_UNICODE} {default_env_name} (runhouse.Env)")
        console.print(
            f"{DOUBLE_SPACE_UNICODE}This environment has only python packages installed, if provided. No "
            "resources were found."
        )

    else:
        # case where the default env have other resources. We make sure that our of all the envs which have resources,
        # the default_env will be printed first.
        first_envs_to_print = [default_env_name]

    # Make sure to print envs with no resources first.
    # (the only resource they have is a runhouse.env, which is the env itself).
    first_envs_to_print = first_envs_to_print + [
        env_name
        for env_name in env_resource_mapping
        if (
            len(env_resource_mapping[env_name]) <= 1
            and env_name != default_env_name
            and env_resource_mapping[env_name]
        )
    ]

    # Now, print the envs.
    # If the env have packages installed, that means that it contains an env resource. In that case:
    # * If the env contains only itself, we will print that the env contains only the installed packages.
    # * Else, we will print the resources (rh.function, th.module) associated with the env.

    envs_to_print = first_envs_to_print + [
        env_name
        for env_name in env_resource_mapping
        if env_name not in first_envs_to_print + [default_env_name]
        and env_resource_mapping[env_name]
    ]

    for env_name in envs_to_print:
        resources_in_env = env_resource_mapping[env_name]
        env_process_info = env_servlet_processes[env_name]

        # sometimes the env itself is not a resource (key) inside the env's servlet.
        if len(resources_in_env) == 0:
            env_type = "runhouse.Env"
        else:
            env_type = _adjust_resource_type(
                resources_in_env[env_name]["resource_type"]
            )

        env_name_txt = f"{BULLET_UNICODE} {env_name} ({env_type}) | pid: {env_process_info['pid']} | node: {env_process_info['node_name']}"
        console.print(env_name_txt)

        # Print CPU info
        env_cpu_info = env_process_info.get("env_cpu_usage")
        if env_cpu_info:

            # convert bytes to GB
            memory_usage_gb = round(
                int(env_cpu_info["used_memory"]) / (1024**3),
                2,
            )
            total_cluster_memory = math.ceil(
                int(env_cpu_info["total_memory"]) / (1024**3)
            )
            cpu_memory_usage_percent = round(
                float(env_cpu_info["used_memory"] / env_cpu_info["total_memory"]),
                2,
            )
            cpu_usage_percent = round(float(env_cpu_info["utilization_percent"]), 2)

            cpu_usage_summery = f"{DOUBLE_SPACE_UNICODE}CPU: {cpu_usage_percent}% | Memory: {memory_usage_gb} / {total_cluster_memory} Gb ({cpu_memory_usage_percent}%)"

        else:
            cpu_usage_summery = (
                f"{DOUBLE_SPACE_UNICODE}CPU: This process did not use CPU memory."
            )

        console.print(cpu_usage_summery)

        # Print GPU info
        env_gpu_info = env_process_info.get("env_gpu_usage")

        # sometimes the cluster has no GPU, therefore the env_gpu_info is an empty dictionary.
        if env_gpu_info:
            # get the gpu usage info, and convert it to GB.
            total_gpu_memory = math.ceil(
                float(env_gpu_info.get("total_memory")) / (1024**3)
            )
            used_gpu_memory = round(
                float(env_gpu_info.get("used_memory")) / (1024**3), 2
            )
            gpu_memory_usage_percent = round(
                float(used_gpu_memory / total_gpu_memory) * 100, 2
            )
            gpu_usage_summery = f"{DOUBLE_SPACE_UNICODE}GPU Memory: {used_gpu_memory} / {total_gpu_memory} Gb ({gpu_memory_usage_percent}%)"
            console.print(gpu_usage_summery)

        resources_in_env = [
            {resource: resources_in_env[resource]}
            for resource in resources_in_env
            if resource is not env_name
        ]

        if len(resources_in_env) == 0:
            # No resources were found in the env, only the associated installed python reqs were installed.
            console.print(
                f"{DOUBLE_SPACE_UNICODE}This environment has only python packages installed, if provided. No resources were "
                "found."
            )

        else:
            for resource in resources_in_env:
                for resource_name, resource_info in resource.items():
                    resource_type = _adjust_resource_type(
                        resource_info.get("resource_type")
                    )

                    active_function_calls = resource_info.get("active_function_calls")
                    resource_info_str = f"{DOUBLE_SPACE_UNICODE}{BULLET_UNICODE} {resource_name} ({resource_type})"

                    if resource_type == "runhouse.Function" and active_function_calls:
                        func_start_time_utc = active_function_calls[0].get(
                            "start_time", None
                        )

                        # casting func_start_time_utc to datetime format
                        func_start_time_utc = datetime.datetime.fromtimestamp(
                            func_start_time_utc, tz=datetime.timezone.utc
                        )

                        # func_end_time_utc = current time. Making sure it is in the same format as func_start_time_utc,
                        # so we could calculate function's running time.
                        func_end_time_utc = datetime.datetime.fromtimestamp(
                            time.time(), tz=datetime.timezone.utc
                        )

                        func_running_time = (
                            func_end_time_utc - func_start_time_utc
                        ).total_seconds()

                        is_func_running: str = f" [italic bright_green]Running for {func_running_time} seconds[/italic bright_green]"

                    elif (
                        resource_type == "runhouse.Function"
                        and not active_function_calls
                    ):
                        is_func_running: str = " [italic bright_yellow]Currently not running[/italic bright_yellow]"

                    else:
                        is_func_running: str = ""

                    resource_info_str = resource_info_str + is_func_running

                    console.print(resource_info_str)


def _print_cloud_properties(cluster_config: dict):
    cloud_properties = cluster_config.get("launched_properties", None)
    if not cloud_properties:
        return
    cloud = cloud_properties.get("cloud")
    instance_type = cloud_properties.get("instance_type")
    region = cloud_properties.get("region")
    cost_per_hour = cloud_properties.get("cost_per_hour")

    has_cuda = cluster_config.get("has_cuda", False)
    cost_emoji = "💰" if has_cuda else "💸"

    num_of_instances = len(cluster_config.get("ips"))
    num_of_instances_str = f"{num_of_instances}x " if num_of_instances > 1 else ""

    print(
        f"🤖 {num_of_instances_str}{cloud} {instance_type} cluster | 🌍 {region} | {cost_emoji} ${cost_per_hour}/hr"
    )


def _get_resource_link_in_den_ui(cluster_name: str, api_server_url: str):
    cluster_uri = rns_client.format_rns_address(cluster_name)
    link_to_den_dashboard = f"{api_server_url}/resources/{cluster_uri}"
    return link_to_den_dashboard


def _print_status(status_data: dict, current_cluster: Cluster) -> None:
    """Prints the status of the cluster to the console"""
    cluster_config = status_data.get("cluster_config")
    env_servlet_processes = status_data.get("env_servlet_processes")

    cluster_name = cluster_config.get("name", None)

    if cluster_name:
        api_server_url = cluster_config.get(
            "api_server_url", rh.configs.get("api_server_url")
        )
        api_server_url = api_server_url.replace(
            "api", "www"
        )  # convert the api link to the ui link.
        cluster_link_in_den_ui = _get_resource_link_in_den_ui(
            cluster_name=cluster_name, api_server_url=api_server_url
        )
        cluster_name_hyperlink = rich.markdown.Text(
            cluster_name, style=f"link {cluster_link_in_den_ui} white"
        )
        console.print(cluster_name_hyperlink)

    has_cuda: bool = cluster_config.get("has_cuda")

    # print headline
    daemon_headline_txt = (
        "\N{smiling face with horns} Runhouse Daemon is running \N{Runner}"
    )
    console.print(daemon_headline_txt, style="bold royal_blue1")

    console.print(f"Runhouse v{status_data.get('runhouse_version')}")
    _print_cloud_properties(cluster_config)
    console.print(f"server pid: {status_data.get('server_pid')}")

    # Print relevant info from cluster config.
    _print_cluster_config(cluster_config)

    # print general cpu and gpu utilization
    cluster_gpu_utilization: float = status_data.get("server_gpu_utilization")

    # cluster_gpu_utilization can be none, if the cluster was not using its GPU at the moment cluster.status() was invoked.
    if cluster_gpu_utilization is None and has_cuda:
        cluster_gpu_utilization: float = 0.0

    cluster_cpu_utilization: float = status_data.get("server_cpu_utilization")

    server_util_info = (
        f"CPU Utilization: {round(cluster_cpu_utilization, 2)}% | GPU Utilization: {round(cluster_gpu_utilization, 2)}%"
        if has_cuda
        else f"CPU Utilization: {round(cluster_cpu_utilization, 2)}%"
    )
    console.print(server_util_info)

    # print the environments in the cluster, and the resources associated with each environment.
    _print_envs_info(env_servlet_processes, current_cluster)


@app.command()
def status(
    cluster_name: str = typer.Argument(
        None,
        help="Name of cluster to check. If not specified will check the local cluster.",
    ),
    send_to_den: bool = typer.Option(
        default=False,
        help="Whether to update Den with the status",
    ),
):
    """Load the status of the Runhouse daemon running on a cluster."""
    cluster_or_local = rh.here

    if cluster_or_local == "file" and not cluster_name:
        # If running outside the cluster must specify a cluster name
        console.print("Missing argument `cluster_name`.")
        return

    elif cluster_name:
        current_cluster = cluster(name=cluster_name)
        if not current_cluster.is_up():
            console.print(
                f"Cluster {cluster_name} is not up. If it's an on-demand cluster, you can run "
                f"`runhouse ssh --up {cluster_name}` to bring it up automatically."
            )
            raise typer.Exit(1)
        try:
            if current_cluster._http_client:
                current_cluster._http_client.check_server()
        except requests.exceptions.ConnectionError:
            console.print(
                f"Could not connect to the server on cluster {cluster_name}. Check that the server is up with "
                f"`runhouse ssh {cluster_name}` or `sky status -r` for on-demand clusters."
            )
            raise typer.Exit(1)

    elif not cluster_or_local:
        console.print(
            "\N{smiling face with horns} Runhouse Daemon is not running... \N{No Entry} \N{Runner}. "
            "Start it with `runhouse restart` or specify a remote "
            "cluster to poll with `runhouse status <cluster_name>`."
        )
        raise typer.Exit(1)

    else:
        # we are inside the cluster
        current_cluster = cluster_or_local  # cluster_or_local = rh.here

    try:
        cluster_status = current_cluster.status(send_to_den=send_to_den)

    except ValueError:
        console.print("Failed to load status for cluster.")
        return
    except requests.exceptions.ConnectionError:
        console.print(
            "\N{smiling face with horns} Runhouse Daemon is not running... \N{No Entry} \N{Runner}"
        )
        return

    _print_status(cluster_status, current_cluster)


def load_cluster(cluster_name: str):
    """Load a cluster from RNS into the local environment, e.g. to be able to ssh."""
    c = cluster(name=cluster_name)
    if not c.address:
        c._update_from_sky_status(dryrun=True)


def _check_if_command_exists(cmd: str):
    cmd_check = subprocess.run(
        f"command -v {cmd}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    available = cmd_check.returncode == 0
    if not available:
        logger.info(f"{cmd} is not available on the system.")
    return available


def _get_wrapped_server_start_cmd(flags: List[str], screen: bool, nohup: bool):
    if screen:
        wrapped_cmd = START_SCREEN_CMD
    elif nohup:
        wrapped_cmd = START_NOHUP_CMD
    else:
        wrapped_cmd = SERVER_START_CMD

    if flags:
        flags_str = "".join(flags)
        wrapped_cmd = wrapped_cmd.replace(
            SERVER_START_CMD, SERVER_START_CMD + flags_str
        )

    return wrapped_cmd


def _start_server(
    restart,
    restart_ray,
    screen,
    nohup,
    create_logfile=True,
    host=None,
    port=None,
    use_https=False,
    den_auth=False,
    ssl_keyfile=None,
    ssl_certfile=None,
    restart_proxy=False,
    use_caddy=False,
    domain=None,
    certs_address=None,
    api_server_url=None,
    default_env_name=None,
    conda_env=None,
    from_python=None,
):
    ############################################
    # Build CLI commands to start the server
    ############################################
    cmds = []
    if restart:
        cmds.append(SERVER_STOP_CMD)

    # We have to `ray start` not within screen/nohup
    existing_ray_instance = check_for_existing_ray_instance()
    if not existing_ray_instance or restart_ray:
        cmds.append(RAY_KILL_CMD)
        cmds.append(RAY_START_CMD)

    # Collect flags
    flags = []

    den_auth_flag = " --use-den-auth" if den_auth else ""
    if den_auth_flag:
        logger.info("Starting server with Den auth.")
        flags.append(den_auth_flag)

    restart_proxy_flag = " --restart-proxy" if restart_proxy else ""
    if restart_proxy_flag:
        logger.info("Reinstalling server configs.")
        flags.append(restart_proxy_flag)

    use_caddy_flag = " --use-caddy" if use_caddy else ""
    if use_caddy_flag:
        logger.info("Configuring Caddy on the cluster.")
        flags.append(use_caddy_flag)

    ssl_keyfile_flag = f" --ssl-keyfile {ssl_keyfile}" if ssl_keyfile else ""
    if ssl_keyfile_flag:
        logger.info(f"Using SSL keyfile in path: {ssl_keyfile}")
        flags.append(ssl_keyfile_flag)

    ssl_certfile_flag = f" --ssl-certfile {ssl_certfile}" if ssl_certfile else ""
    if ssl_certfile_flag:
        logger.info(f"Using SSL certfile in path: {ssl_certfile}")
        flags.append(ssl_certfile_flag)

    domain = f" --domain {domain}" if domain else ""
    if domain:
        logger.info(f"Using domain: {domain}")
        flags.append(domain)

    # Use HTTPS if explicitly specified or if SSL cert or keyfile path are provided
    https_flag = " --use-https" if use_https or (ssl_keyfile or ssl_certfile) else ""
    if https_flag:
        logger.info("Starting server with HTTPS.")
        flags.append(https_flag)

    host_flag = f" --host {host}" if host else ""
    if host_flag:
        logger.info(f"Using host: {host}.")
        flags.append(host_flag)

    port_flag = f" --port {port}" if port else ""
    if port_flag:
        logger.info(f"Using port: {port}.")
        flags.append(port_flag)

    address_flag = f" --certs-address {certs_address}" if certs_address else ""
    if address_flag:
        logger.info(f"Server public IP address: {certs_address}.")
        flags.append(address_flag)

    api_server_url_flag = (
        f" --api-server-url {api_server_url}" if api_server_url else ""
    )
    if api_server_url_flag:
        logger.info(f"Setting api_server url to {api_server_url}")
        flags.append(api_server_url_flag)

    default_env_flag = (
        f" --default-env-name {default_env_name}" if default_env_name else ""
    )
    if default_env_flag:
        logger.info(f"Starting server in default env named: {default_env_name}")
        flags.append(default_env_flag)

    conda_env_flag = f" --conda-env {conda_env}" if conda_env else ""
    if conda_env_flag:
        logger.info(f"Creating runtime env for conda env: {conda_env}")
        flags.append(conda_env_flag)

    flags.append(" --from-python" if from_python else "")

    # Check if screen or nohup are available
    screen = screen and _check_if_command_exists("screen")
    nohup = not screen and nohup and _check_if_command_exists("nohup")

    # Create logfile if we are using backgrounding
    if (screen or nohup) and create_logfile and not Path(SERVER_LOGFILE).exists():
        Path(SERVER_LOGFILE).parent.mkdir(parents=True, exist_ok=True)
        Path(SERVER_LOGFILE).touch()

    # Add flags to the server start command
    cmds.append(_get_wrapped_server_start_cmd(flags, screen, nohup))
    logger.info(f"Starting API server using the following command: {cmds[-1]}.")

    try:
        # Open and read the lines of the server logfile so we only print the most recent lines after starting
        f = None
        if (screen or nohup) and Path(SERVER_LOGFILE).exists():
            f = open(SERVER_LOGFILE, "r")
            f.readlines()  # Discard these, they're from the previous times the server was started

        # We do these one by one so it's more obvious where the error is if there is one
        for i, cmd in enumerate(cmds):
            console.print(f"Executing `{cmd}`")
            if (
                i == len(cmds) - 1
            ):  # last cmd is not being parsed correctly when ran with shlex.split
                result = subprocess.run(cmd, shell=True, check=True)
            else:
                result = subprocess.run(shlex.split(cmd), text=True)

            if result.returncode != 0:
                # We don't want to raise an error if the server kill fails, as it may simply not be running
                if "pkill" in cmd:
                    continue

                # Retry ray start in case pkill process did not complete in time, up to 10s
                if cmd == RAY_START_CMD:
                    console.print("Retrying:")
                    attempt = 0
                    while result.returncode != 0 and attempt < 10:
                        attempt += 1
                        time.sleep(1)
                        result = subprocess.run(
                            shlex.split(cmd), text=True, capture_output=True
                        )
                        if result.stderr and "ConnectionError" not in result.stderr:
                            break
                    if result.returncode == 0:
                        continue

                console.print(f"Error while executing `{cmd}`")
                raise typer.Exit(1)

        server_started_str = "Uvicorn running on"
        # Read and print the server logs until the
        if screen or nohup:
            while not Path(SERVER_LOGFILE).exists():
                time.sleep(1)
            f = f or open(SERVER_LOGFILE, "r")
            start_time = time.time()
            # Wait for input for 60 seconds max (for Caddy to download and set up)
            while time.time() - start_time < 60:
                for line in f:
                    if server_started_str in line:
                        console.print(line)
                        f.close()
                        return
                    else:
                        console.print(line, end="")
                time.sleep(1)
            f.close()

    except FileNotFoundError:
        console.print(
            "python3 command was not found. Make sure you have python3 installed."
        )
        raise typer.Exit(1)


@app.command()
def start(
    restart_ray: bool = typer.Option(True, help="Restart the Ray runtime"),
    screen: bool = typer.Option(False, help="Start the server in a screen"),
    nohup: bool = typer.Option(
        False, help="Start the server in a nohup if screen is not available"
    ),
    host: Optional[str] = typer.Option(
        None, help="Custom server host address. Default is `0.0.0.0`."
    ),
    port: Optional[str] = typer.Option(
        None, help="Port for server. If not specified will start on 32300"
    ),
    use_https: bool = typer.Option(
        False, help="Start an HTTPS server with TLS verification"
    ),
    use_den_auth: bool = typer.Option(
        False, help="Whether to authenticate requests with a Runhouse token"
    ),
    use_caddy: bool = typer.Option(
        False,
        help="Whether to configure Caddy on the cluster as a reverse proxy.",
    ),
    domain: str = typer.Option(
        None,
        help="Server domain. Relevant if using Caddy to automate generating CA verified certs.",
    ),
    certs_address: Optional[str] = typer.Option(
        None,
        help="Public IP address of the server. Required for generating self-signed certs and enabling HTTPS",
    ),
    default_env_name: str = typer.Option(
        None, help="Default env to start the server on."
    ),
    conda_env: str = typer.Option(
        None, help="Name of conda env corresponding to default env if it is a CondaEnv."
    ),
):
    """Start the HTTP or HTTPS server on the cluster."""
    _start_server(
        restart=False,
        restart_ray=restart_ray,
        screen=screen,
        nohup=nohup,
        create_logfile=True,
        host=host,
        port=port,
        use_https=use_https,
        den_auth=use_den_auth,
        use_caddy=use_caddy,
        domain=domain,
        certs_address=certs_address,
        default_env_name=default_env_name,
        conda_env=conda_env,
    )


@app.command()
def restart(
    name: str = typer.Option(None, help="A *saved* remote cluster object to restart."),
    restart_ray: bool = typer.Option(True, help="Restart the Ray runtime"),
    screen: bool = typer.Option(
        True,
        help="Start the server in a screen. Only relevant when restarting locally.",
    ),
    nohup: bool = typer.Option(
        True,
        help="Start the server in a nohup if screen is not available. Only relevant when restarting locally.",
    ),
    resync_rh: bool = typer.Option(
        False,
        help="Resync the Runhouse package. Only relevant when restarting remotely.",
    ),
    host: Optional[str] = typer.Option(
        None, help="Custom server host address. Default is `0.0.0.0`."
    ),
    port: Optional[str] = typer.Option(
        None, help="Port for server. If not specified will start on 32300"
    ),
    use_https: bool = typer.Option(
        False, help="Start an HTTPS server with TLS verification"
    ),
    use_den_auth: bool = typer.Option(
        False, help="Whether to authenticate requests with a Runhouse token"
    ),
    ssl_keyfile: Optional[str] = typer.Option(
        None, help="Path to custom SSL key file to use for enabling HTTPS"
    ),
    ssl_certfile: Optional[str] = typer.Option(
        None, help="Path to custom SSL cert file to use for enabling HTTPS"
    ),
    restart_proxy: bool = typer.Option(
        False, help="Whether to reinstall server configs on the cluster"
    ),
    use_caddy: bool = typer.Option(
        False,
        help="Whether to configure Caddy on the cluster as a reverse proxy.",
    ),
    domain: str = typer.Option(
        None,
        help="Server domain. Relevant if using Caddy to automate generating CA verified certs.",
    ),
    certs_address: Optional[str] = typer.Option(
        None,
        help="Public IP address of the server. Required for generating self-signed certs and enabling HTTPS",
    ),
    api_server_url: str = typer.Option(
        default="https://api.run.house",
        help="URL of Runhouse Den",
    ),
    default_env_name: str = typer.Option(
        None, help="Default env to start the server on."
    ),
    conda_env: str = typer.Option(
        None, help="Name of conda env corresponding to default env if it is a CondaEnv."
    ),
    from_python: bool = typer.Option(
        False,
        help="Whether HTTP server started from inside a Python call rather than CLI.",
    ),
):
    """Restart the HTTP server on the cluster."""
    if name:
        c = cluster(name=name)
        c.restart_server(resync_rh=resync_rh, restart_ray=restart_ray)
        return

    _start_server(
        restart=True,
        restart_ray=restart_ray,
        screen=screen,
        nohup=nohup,
        create_logfile=True,
        host=host,
        port=port,
        use_https=use_https,
        den_auth=use_den_auth,
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile,
        restart_proxy=restart_proxy,
        use_caddy=use_caddy,
        domain=domain,
        certs_address=certs_address,
        api_server_url=api_server_url,
        default_env_name=default_env_name,
        conda_env=conda_env,
        from_python=from_python,
    )


@app.command()
def stop(
    stop_ray: bool = typer.Option(False, help="Stop the Ray runtime"),
    cleanup_actors: bool = typer.Option(True, help="Kill all Ray actors"),
):
    logger.info("Stopping the server.")
    subprocess.run(SERVER_STOP_CMD, shell=True)

    if cleanup_actors:
        ray.init(
            address="auto",
            ignore_reinit_error=True,
            logging_level=logging.ERROR,
            namespace="runhouse",
        )

        kill_actors(namespace="runhouse", gracefully=False)

    if stop_ray:
        logger.info("Stopping Ray.")
        subprocess.run(RAY_KILL_CMD, shell=True)


@app.callback(invoke_without_command=True, help="Runhouse CLI")
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", "-v", help="Show the version and exit."
    ),
):
    """
    Runhouse CLI
    """
    if version:
        print(f"{__version__}")
    elif ctx.invoked_subcommand is None:
        subprocess.run("runhouse --help", shell=True)



import importlib
from pathlib import Path

import pytest
import runhouse as rh

from tests.utils import get_pid_and_ray_node


class TestMultiNodeCluster:
    @pytest.mark.level("release")
    def test_rsync_and_ssh_onto_worker_node(self, multinode_cpu_docker_conda_cluster):
        worker_node = multinode_cpu_docker_conda_cluster.ips[-1]
        local_rh_package_path = Path(importlib.util.find_spec("runhouse").origin).parent

        local_rh_package_path = local_rh_package_path.parent
        dest_path = f"~/{local_rh_package_path.name}"

        # Rsync Runhouse package onto the worker node
        multinode_cpu_docker_conda_cluster.rsync(
            source=str(local_rh_package_path),
            dest=dest_path,
            up=True,
            node=worker_node,
            contents=True,
        )

        status_codes = multinode_cpu_docker_conda_cluster.run(
            [f"ls -l {dest_path}"], node=worker_node
        )
        assert status_codes[0][0] == 0

        assert "runhouse" in status_codes[0][1]

    @pytest.mark.level("release")
    def test_ray_started_on_worker_node_after_cluster_restart(
        self, multinode_cpu_docker_conda_cluster
    ):
        head_node = multinode_cpu_docker_conda_cluster.ips[0]

        status_codes = multinode_cpu_docker_conda_cluster.run(
            ["ray status"], node=head_node
        )
        assert status_codes[0][0] == 0

        status_output = status_codes[0][1]
        node_marker = "1 node_"
        num_nodes = status_output.count(node_marker)
        assert num_nodes == 2

    @pytest.mark.level("release")
    def test_send_envs_to_specific_worker_node(
        self, multinode_cpu_docker_conda_cluster
    ):

        env_0 = rh.env(
            name="worker_env_0",
            reqs=["langchain", "pytest"],
        ).to(multinode_cpu_docker_conda_cluster, node_idx=0)

        env_1 = rh.env(
            name="worker_env_1",
            reqs=["torch", "pytest"],
        ).to(multinode_cpu_docker_conda_cluster, node_idx=1)

        env_2 = rh.env(
            name="worker_env_2",
            reqs=["transformers", "pytest"],
        )

        with pytest.raises(ValueError):
            env_2.to(
                multinode_cpu_docker_conda_cluster,
                node_idx=len(multinode_cpu_docker_conda_cluster.ips),
            )

        env_2.to(multinode_cpu_docker_conda_cluster, node_idx=1)

        get_pid_0 = rh.function(get_pid_and_ray_node).to(
            name="get_pid_0", system=multinode_cpu_docker_conda_cluster, env=env_0
        )
        get_pid_1 = rh.function(get_pid_and_ray_node).to(
            name="get_pid_1", system=multinode_cpu_docker_conda_cluster, env=env_1
        )
        get_pid_2 = rh.function(get_pid_and_ray_node).to(
            name="get_pid_2", system=multinode_cpu_docker_conda_cluster, env=env_2
        )
        assert get_pid_0()[1] != get_pid_1()[1]
        assert get_pid_1()[1] == get_pid_2()[1]

    @pytest.mark.level("release")
    def test_specifying_resources(self, multinode_cpu_docker_conda_cluster):
        env0 = rh.env(
            name="worker_env_0",
            compute={"CPU": 1.75},
        ).to(multinode_cpu_docker_conda_cluster)

        env1 = rh.env(
            name="worker_env_1",
            compute={"CPU": 0.5},
        ).to(multinode_cpu_docker_conda_cluster)

        env2 = rh.env(
            name="worker_env_2",
            compute={"memory": 4 * 1024 * 1024 * 1024},
        ).to(multinode_cpu_docker_conda_cluster)

        env3 = rh.env(
            name="worker_env_3",
            compute={"CPU": 0.1, "memory": 2 * 1024 * 1024 * 1024},
        ).to(multinode_cpu_docker_conda_cluster)

        status = multinode_cpu_docker_conda_cluster.status()

        env0_node = status["env_servlet_processes"][env0.name]["node_ip"]
        env1_node = status["env_servlet_processes"][env1.name]["node_ip"]
        env2_node = status["env_servlet_processes"][env2.name]["node_ip"]
        env3_node = status["env_servlet_processes"][env3.name]["node_ip"]
        assert env0_node in multinode_cpu_docker_conda_cluster.internal_ips
        assert env1_node in multinode_cpu_docker_conda_cluster.internal_ips
        assert env2_node in multinode_cpu_docker_conda_cluster.internal_ips
        assert env3_node in multinode_cpu_docker_conda_cluster.internal_ips

        assert env0_node != env1_node  # Too much CPU
        assert env2_node != env3_node  # Too much memory



from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Union

from runhouse.resources.packages import Package

from .conda_env import CondaEnv

from .env import Env
from .utils import _get_conda_yaml, _process_reqs


# generic Env factory method
def env(
    reqs: List[Union[str, Package]] = [],
    conda_env: Union[str, Dict] = None,
    name: Optional[str] = None,
    setup_cmds: List[str] = None,
    env_vars: Union[Dict, str] = {},
    working_dir: Optional[Union[str, Path]] = None,
    secrets: Optional[Union[str, "Secret"]] = [],
    compute: Optional[Dict] = {},
    load_from_den: bool = True,
    dryrun: bool = False,
):
    """Builds an instance of :class:`Env`.

    Args:
        reqs (List[str]): List of package names to install in this environment.
        conda_env (Union[str, Dict], optional): Dict representing conda env, Path to a conda env yaml file,
            or name of a local conda environment.
        name (Optional[str], optional): Name of the environment resource.
        setup_cmds (Optional[List[str]]): List of CLI commands to run for setup when the environment is
            being set up on a cluster.
        env_vars (Dict or str): Dictionary of environment variables, or relative path to .env file containing
            environment variables. (Default: {})
        working_dir (str or Path): Working directory of the environment, to be loaded onto the system.
            (Default: None)
        compute (Dict): Logical compute resources to be used by this environment, passed through to the
            cluster scheduler (generally Ray). Only use this if you know what you're doing.
            Example: ``{"cpus": 1, "gpus": 1}``. (Default: {})
            More info: https://docs.ray.io/en/latest/ray-core/scheduling/resources.html
        load_from_den (bool): Whether to try loading the Env resource from Den. (Default: ``True``)
        dryrun (bool, optional): Whether to run in dryrun mode. (Default: ``False``)


    Returns:
        Env: The resulting Env object.

    Example:
        >>> # regular python env
        >>> env = rh.env(reqs=["torch", "pip"])
        >>> env = rh.env(reqs=["reqs:./"], name="myenv")
        >>>
        >>> # conda env, see also rh.conda_env
        >>> conda_env_dict =
        >>>     {"name": "new-conda-env", "channels": ["defaults"], "dependencies": "pip", {"pip": "diffusers"})
        >>> conda_env = rh.env(conda_env=conda_env_dict)             # from a dict
        >>> conda_env = rh.env(conda_env="conda_env.yaml")           # from a yaml file
        >>> conda_env = rh.env(conda_env="local-conda-env-name")     # from a existing local conda env
        >>> conda_env = rh.env(conda_env="conda_env.yaml", reqs=["pip:/accelerate"])   # with additional reqs
    """
    if name and not any(
        [reqs, conda_env, setup_cmds, env_vars, secrets, working_dir, compute]
    ):
        try:
            return Env.from_name(name, load_from_den=load_from_den, dryrun=dryrun)
        except ValueError:
            return Env(name=name)

    if not name and compute:
        raise ValueError("Cannot specify compute to schedule an env on without a name.")

    reqs = _process_reqs(reqs or [])
    conda_yaml = _get_conda_yaml(conda_env)

    if conda_yaml:
        return CondaEnv(
            conda_yaml=conda_yaml,
            reqs=reqs,
            setup_cmds=setup_cmds,
            env_vars=env_vars,
            working_dir=working_dir,
            secrets=secrets,
            name=name or conda_yaml["name"],
            dryrun=dryrun,
        )

    return Env(
        reqs=reqs,
        setup_cmds=setup_cmds,
        env_vars=env_vars,
        working_dir=working_dir,
        secrets=secrets,
        name=name,
        compute=compute,
        dryrun=dryrun,
    )


# Conda Env factory method
def conda_env(
    reqs: List[Union[str, Package]] = [],
    conda_env: Union[str, Dict] = None,
    name: Optional[str] = None,
    setup_cmds: List[str] = None,
    env_vars: Optional[Dict] = {},
    working_dir: Optional[Union[str, Path]] = None,
    secrets: List[Union[str, "Secret"]] = [],
    compute: Optional[Dict] = {},
    dryrun: bool = False,
):
    """Builds an instance of :class:`CondaEnv`.

    Args:
        reqs (List[str]): List of package names to install in this environment.
        conda_env (Union[str, Dict], optional): Dict representing conda env, Path to a conda env yaml file,
            or name of a local conda environment.
        name (Optional[str], optional): Name of the environment resource.
        setup_cmds (Optional[List[str]]): List of CLI commands to run for setup when the environment is
            being set up on a cluster.
        env_vars (Dict or str): Dictionary of environment variables, or relative path to .env file containing
            environment variables. (Default: {})
        working_dir (str or Path): Working directory of the environment, to be loaded onto the system.
            (Default: None)
        compute (Dict): Logical compute resources to be used by this environment, passed through to the
            cluster scheduler (generally Ray). Only use this if you know what you're doing.
            Example: ``{"cpus": 1, "gpus": 1}``. (Default: {})
            More info: https://docs.ray.io/en/latest/ray-core/scheduling/resources.html
        dryrun (bool, optional): Whether to run in dryrun mode. (Default: ``False``)

    Returns:
        CondaEnv: The resulting CondaEnv object.

    Example:
        >>> rh.conda_env(reqs=["torch"])
        >>> rh.conda_env(reqs=["torch"], name="resource_name")
        >>> rh.conda_env(reqs=["torch"], name="resource_name", conda_env={"name": "conda_env"})
    """
    if not conda_env:
        if name:
            conda_env = {"name": name}
        else:
            conda_env = {"name": datetime.now().strftime("%Y%m%d_%H%M%S")}

    return env(
        reqs=reqs,
        conda_env=conda_env,
        name=name,
        setup_cmds=setup_cmds,
        env_vars=env_vars,
        working_dir=working_dir,
        secrets=secrets,
        compute=compute,
        dryrun=dryrun,
    )
